#!/usr/bin/env python
# coding: utf-8

import argparse
import configparser
import pathlib as pl

from spifpy.input import DMTMonoFile
from spifpy.input import DMTGreyFile
from spifpy.input import SPECFile
from spifpy.input import DMTGreyMonoFile
from spifpy.input import TwoDFile
from spifpy.spif import SPIFCore

inst_dict = {'2DC': TwoDFile,
             '2DP': TwoDFile,
             'CIP': DMTMonoFile,
             'CIPGS': DMTGreyFile,
             'PIP': DMTMonoFile,
             '2DS': SPECFile,
             'HVPS': SPECFile,
             'HVPS4': SPECFile}

def extract():

     # Get the parser and grab the args

     parser = get_parser()
     args = parser.parse_args()

     # Check the args, make sure everything is OK before using
     # the arguments to do anything with them

     args_checker = ArgsChecker(args)
     args_checker.check_args()

     # Transform the arguments to the proper forms for smooth processing

     args_transformer = ArgsTransformer(args)
     transformed_args = args_transformer.transform_args()

     spif_core = call_spifcore(transformed_args)
     spif_core.process(processors=None)

def get_parser():
    parser = argparse.ArgumentParser(description='Processes raw OAP data to' +
                                                 ' SPIF formatted NetCDF file')
    parser.add_argument('filename',
                        type=str,
                        help= 'path to raw instrument file to process.')

    parser.add_argument('config',
                        type=str,
                        help= 'path to config file to use for processing')

    parser.add_argument('-o',
                        dest='output',
                        type=str,
                        help='Filename to use for SPIF output',
                        default=None)

    return parser

def call_spifcore(transformed_args):
     inst_name, inst_class = get_inst_name(transformed_args)
     filename = transformed_args['filename']
     outfile = transformed_args['output']
     config = transformed_args['config']

     spif_core = SPIFCore(
          inst_class,
          filename,
          outfile,
          config
     )

     return spif_core

def get_inst_name(transformed_args):
     config = configparser.ConfigParser(allow_no_value=True)

     config.read(transformed_args['config'])
     inst_name = config['instrument'].get('instrument_name', None)
     inst_class = inst_dict[inst_name]
     if config['instrument'].getboolean('mono_as_grey', False):
          inst_class = DMTGreyMonoFile

     return inst_name, inst_class

class ArgsChecker:
     def __init__(self, args):
          self.args = args

     def check_args(self):
          self.check_filename(self.args.filename)
          self.check_config(self.args.config)

          if self.args.output is not None : self.check_output(self.args.output)

     def check_filename(self, filename):

          try:
               assert pl.Path(filename).is_file()
          except AssertionError:
               print(f"ERROR : The file {filename} is not a valid file")
               raise

          try:
               assert pl.Path(filename).stat().st_size
          except AssertionError:
               print(f"ERROR : The file {filename} is an empty file")
               raise

     def check_config(self, config_file):
          config = configparser.ConfigParser(allow_no_value=True)

          config.read(config_file)
          inst_name = config['instrument'].get('instrument_name', None)

          try:
               assert inst_name in inst_dict
          except AssertionError:
               print(f"ERROR : The provided instrument type {inst_name} from config file {config} is invalid. Please provide a valid instrument from the following list {' | '.join([k for k in inst_dict.keys()])}")
               raise

     def check_output(self, output):

          try:
               open(output, 'w')
          except OSError:
               print(f"ERROR : The provided output filename {output} is not a valid filename.")
               raise

          try:
               assert pl.Path(output).suffix == '.nc'
          except AssertionError:
               print(f"ERROR : The provided filename {output} does not have a .nc file ending")
               raise

class ArgsTransformer:

     def __init__(self, args) -> None:
         self.args = args

         self.transformed_args = {
              'filename':None,
              'config':None,
              'output':None,
              'aux_file':None,
              'aux_config':None,
         }

     def transform_args(self):
          self.transform_filename()
          self.transform_config()
          self.transform_output()

          return self.transformed_args

     def transform_filename(self):
          self.transformed_args['filename'] = pl.Path(self.args.filename)

     def transform_config(self):
          self.transformed_args['config'] = self.args.config

     def transform_output(self):
          if self.args.output == None:
               input = pl.Path(self.args.filename)
               self.transformed_args['output'] = input.parent / (input.name.replace('.','_') + '.nc')
          else:
               self.transformed_args['output'] = pl.Path(self.args.output)



import logging
import os
import sys
from datetime import datetime, timedelta
from os import PathLike
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Union

import numpy as np
import numpy.typing as npt
import xarray as xr

from spifpy.probes import get_probe, SpifProbe

logger = logging.getLogger(__name__)


class SpifDatasetException(Exception):
    pass


class ParticleData(NamedTuple):
    """Typing for SPIF particle data."""

    image_id: int
    image: npt.NDArray
    timestamp: datetime
    metadata: Optional[Dict[str, Any]]


ParticleDataList = List[ParticleData]
FilePath = Union[str, PathLike]


class SpifDataset:
    """Generic SPIF dataset class."""

    SUPPORTED_PROBES = ["2DC", "2DS-H", "2DS-V", "CIP", "CPI", "HSI", "HVPS", "PIP"]
    DEFAULT_TIME_BASIS = "seconds since start_date"
    DEFAULT_TIME_FORMAT = "%F %T %z"

    def __init__(self, fpath: FilePath, probe: Optional[str] = None):
        self.fpath = fpath

        if not os.path.exists(self.fpath) or not os.path.isfile(self.fpath):
            raise FileNotFoundError(
                f"Spif Dataset file not found under path {self.fpath}"
            )

        self.probe: SpifProbe = self._read_probe(probe)
        self._init_dataset()

    def _init_dataset(self):
        self.probe_group = self._get_probe_group()
        self.core_group = self._get_core_group()
        self.start_date = self._get_start_date()
        self.data_dimension = self._get_data_dimension()
        self.image_widths = self._get_image_widths()
        self.image_heights = self._get_image_heights()
        self.image_indices = self._get_indices()

        self.datetimes = self._convert_datetimes()
        self.start_index, self.end_index = 0, len(self.datetimes) - 1

    def _convert_datetimes(self):
        particle_sec = self.core_group[self._get_seconds_variable_name()]
        particle_ns = self.core_group[self._get_nanoseconds_variable_name()]
        time_data = particle_sec[:] + particle_ns[:] * 1e-9

        particle_time_format = SpifDataset.DEFAULT_TIME_FORMAT
        if "strftime_format" in particle_sec.attrs:
            particle_time_format = particle_sec.strftime_format
        elif "strptime_format" in particle_sec.attrs:
            # Deprecated: will be removed in the future release because
            # 'strptime_format' has been renamed to 'strftime_format'
            particle_time_format = particle_sec.strptime_format

        time_basis = (
            particle_sec.units
            if "units" in particle_sec.attrs
            else SpifDataset.DEFAULT_TIME_BASIS
        )

        datetimes = self.convert_netcdf_time(
            time_data.data,
            time_basis,
            particle_time_format,
            start_date=self.start_date,
        )

        return datetimes

    def _get_core_group(self):
        return xr.open_dataset(
            self.fpath,
            group=f"{self.probe.get_name()}/core/",
            engine="h5netcdf",
            decode_times=False,
        )

    def _get_data_dimension(self):
        return self.core_group[self._get_image_data_field_name()].ndim

    def _get_image_data_field_name(self):
        return "image"

    def _get_image_heights(self):
        return np.repeat(
            self.probe_group["pixels"].data, self.image_widths.shape[0]
        ).astype(np.uint32)

    def _get_image_widths(self):
        return self.core_group["image_len"].data.astype(np.uint32)

    def _get_images_axis_name(self):
        return "Images"

    def _get_images_batch(self, batch_start: int, batch_end: int):
        if self.data_dimension == 3:
            return self.core_group[
                {self._get_images_axis_name(): slice(batch_start, batch_end)}
            ]
        elif self.data_dimension == 1:
            return self.core_group[
                {
                    self._get_pixels_axis_name(): slice(
                        self.image_indices[batch_start],
                        self.image_indices[batch_end],
                    )
                }
            ]
        else:
            raise SpifDatasetException("2D image arrays are not supported")

    def _get_pixels_axis_name(self):
        return "Pixels"

    def _get_seconds_variable_name(self):
        return "image_sec"

    def _get_nanoseconds_variable_name(self):
        return "image_ns"

    def _get_indices(self):
        indices = np.cumsum(self.image_widths * self.image_heights)
        indices = np.insert(indices, 0, 0)
        return indices

    def _get_probe_group(self):
        return xr.open_dataset(
            self.fpath,
            engine="h5netcdf",
            group=self.probe.get_name(),
            decode_times=False,
        )

    def _get_start_date(self):
        return xr.open_dataset(self.fpath, engine="h5netcdf").start_date.strip()

    def _process_batch(self, batch, *args, **kwargs):
        if self.data_dimension == 3:
            return self._process_batch_3d(batch, *args, **kwargs)
        elif self.data_dimension == 1:
            return self._process_batch_1d(batch, *args, **kwargs)
        else:
            raise SpifDatasetException("2D image arrays are not supported")

    def _process_batch_3d(self, batch, *args, **kwargs):
        batch_start, batch_end = args

        images = batch[self._get_image_data_field_name()].data
        widths = self.image_widths[batch_start:batch_end]
        heights = self.image_heights[batch_start:batch_end]
        datetimes = self.datetimes[batch_start:batch_end]

        result: ParticleDataList = []

        for image_num, (
            image,
            image_width,
            image_height,
            timestamp,
        ) in enumerate(zip(images, widths, heights, datetimes)):
            image_id = batch_start + image_num
            image = image[:image_width, :]
            image = image.reshape((-1, image_height)).T

            result.append(ParticleData(image_id, image, timestamp, None))

        return result

    def _process_batch_1d(self, batch, *args, **kwargs):
        batch_start, batch_end = args

        images = batch[self._get_image_data_field_name()].data
        widths = self.image_widths[batch_start:batch_end]
        heights = self.image_heights[batch_start:batch_end]
        datetimes = self.datetimes[batch_start:batch_end]

        batch_indices = np.cumsum(widths * heights)
        batch_indices = np.insert(batch_indices, 0, 0)

        result: ParticleDataList = []

        for image_num, timestamp in enumerate(datetimes):
            image_id = batch_start + image_num
            if batch_indices[image_num + 1] > len(images):
                raise SpifDatasetException(
                    "Not enough pixel data: number of elements "
                    + "in a pixels array do not equal to multiplication "
                    + "of image widths by heights."
                )

            image = images[batch_indices[image_num] : batch_indices[image_num + 1]]
            if image.shape != (0,):
                image = image.reshape((widths[image_num], heights[image_num])).T

            result.append(ParticleData(image_id, image, timestamp, None))

        return result

    def _read_probe(self, probe: Optional[str] = None) -> SpifProbe:
        available_probes = self.get_available_probes()

        if not probe:
            if len(available_probes) > 1:
                logger.warn(
                    "More than 1 probe found in the dataset file. "
                    f"By default the first probe {available_probes[0]} will be used. "
                    "To change that, specify probe parameter when calling SpifDataset constructor."
                )

            if len(available_probes) >= 1:
                probe_name = available_probes[0]
                return get_probe(probe_name)

            raise SpifDatasetException(
                "No probe groups found in the dataset file. "
                "The file is likely corrupted."
            )

        return get_probe(probe)

    @staticmethod
    def convert_netcdf_time(time_data, time_basis, basis_format, start_date):
        """
        Convert time in standard NetCDF time format into datetime object.
        Author: Matt Freer (mfreer@gmail.com)
        Parameters
        ----------
        time_data : array
            Timestamps of data contained within the file,
            seconds since `start_date`
        time_units : str
            Time units string in the format: 'seconds since start_date'
            `start_date` string is then replaced with the start date parameter
        basis_format: str
            Time format string, e.g. '%F %T %z'
        start_date: str
            Start date timestamp in a string format
        Returns
        -------
        list of datetime object
            Absolute datetimes of provided times.
        """
        TIME_UNITS_DEFAULT = "seconds since "
        basis_format = basis_format.replace("%F", "%Y-%m-%d")
        basis_format = basis_format.replace("%T", "%H:%M:%S")

        if not basis_format.startswith(TIME_UNITS_DEFAULT):
            basis_format = TIME_UNITS_DEFAULT + basis_format

        time_basis = time_basis.replace("start_date", start_date)
        try:
            tb_dt = datetime.strptime(time_basis, basis_format)
        except ValueError:
            tb_dt = datetime.strptime(
                time_basis, basis_format.replace("%z", "").strip()
            )
        time_dt = [tb_dt + timedelta(seconds=t) for t in time_data]

        return time_dt

    def get_available_probes(self) -> List[str]:
        """
        Based on the file, return the list of its probes.

        Returns
        -------
        List[str]
            List of probes, ordered alphabetically,
            data of which this file contains
        """
        file_probes = []

        for probe in SpifDataset.SUPPORTED_PROBES:
            try:
                dataset = xr.open_dataset(
                    self.fpath,
                    engine="h5netcdf",
                    group=f"{probe}/core/",
                    decode_times=False,
                )
                dataset.close()
                file_probes.append(probe)
            except Exception:  # nosec
                # we shouldn't ideally continue after exception,
                # potentially look for a better solution to check available
                # probes in the file
                continue

        return sorted(file_probes)

    def get_end_time(self) -> datetime:
        """Get timestamp of the last image in the dataset."""
        return self.datetimes[-1]

    def get_file_path(self) -> FilePath:
        """Get dataset file path."""
        return self.fpath

    def get_image_id_by_timestamp(self, timestamp: datetime) -> int:
        """
        Get image index by timestamp.

        Considering that all timestamps are ordered in the dataset,
        look up an image that has a timestamp closest to the passed value.

        If the passed timestamp is earlier than the timestamp of the first image,
        0 will be returned. If it is later than the timestamp of the last image,
        the index of the last image will be returned.
        """
        if len(self.datetimes) == 0:
            return 0

        if timestamp > self.datetimes[-1]:
            return len(self.datetimes) - 1

        if timestamp < self.datetimes[0]:
            return 0

        for i, array_timestamp in enumerate(self.datetimes):
            if array_timestamp > timestamp:
                return i

        return 0

    def get_images_number(self) -> int:
        """Get number of images in the dataset."""
        return len(self.datetimes)

    def get_probe(self) -> SpifProbe:
        """Get an instance of the currently open probe."""
        return self.probe

    def get_start_time(self) -> datetime:
        """Get timestamp of the first image in the dataset."""
        return self.datetimes[0]

    def read(self, start: int = 0, end: int = sys.maxsize) -> Iterator[ParticleData]:
        start_index = max(self.start_index, start)
        end_index = min(self.end_index, end)

        for batch_start in range(start_index, end_index):
            batch_end = batch_start + 1

            if batch_end > end_index + 1:
                batch_end = end_index + 1

            batch = self._get_images_batch(batch_start, batch_end)
            batch = self._process_batch(batch, batch_start, batch_end)

            yield batch[0]

    def read_batch(
        self, start: int = 0, end: int = sys.maxsize, batch_size=1
    ) -> Iterator[ParticleDataList]:
        start_index = max(self.start_index, start)
        end_index = min(self.end_index, end)

        for batch_start in range(start_index, end_index, batch_size):
            batch_end = batch_start + batch_size

            if batch_end > end_index + 1:
                batch_end = end_index + 1

            batch = self._get_images_batch(batch_start, batch_end)
            batch = self._process_batch(batch, batch_start, batch_end)

            yield batch



#!/usr/bin/env python
# coding: utf-8

import math

from numba import njit
import numpy


def convert_datetimes_to_seconds(start_date, datetimes):
    """ Converts difference in datetimes to total elapsed seconds.

    Parameters
    ----------
    start_date : datetime object
        Start date to use for calculating total elapsed seconds.
    datetimes : list of datetimes
        List of datetimes to calculate total elapsed seconds for.

    Returns
    -------
    list of float
        Total elapsed seconds from start_date for each element in datetimes list.
    """
    datetimes_seconds = [(dt - start_date).total_seconds() for dt in datetimes]

    return datetimes_seconds


def bin_to_int(bin_list, inverse=True):
    """ Given arbitrary length list of binary values, calculate integer value.

    Parameters
    ----------
    bin_list : list of int
        List of binary values for calculating integer value.
    inverse : bool, optional
        If true (default) caluclate integer from bin_list assuming first element
        is LSB. If false, integer will be calculated from bin_list assuming first
        element is MSB.

    Returns
    -------
    int
        Integer value calculated from bin_list.
    """
    if inverse:
        step = -1
    else:
        step = 1
    bin_list = bin_list[::step]
    bin_int = 0
    for bit in bin_list:
        bin_int = (bin_int << 1) | bit
    return int(bin_int)


def list_to_array(lst, diodes):
    """ Convert list of ragged length 1-D arrays to square numpy array.

    Parameters
    ----------
    lst : list of 1-D arrays
        List of arrays to normalize
    diodes : int
        Number of diodes in instrument array.

    Returns
    -------
    array
        Square 2-d array of input arrays.
    int
        Maximum length of any individual 1-D array, scaled to next factor
        of n-diodes to allow reshaping at later step.
    """
    lengths = numpy.array([len(item) for item in lst])
    max_length = math.ceil(lengths.max() / diodes) * diodes
    mask = lengths[:, None] > numpy.arange(max_length)
    out = numpy.zeros(mask.shape, dtype='uint8')
    out[mask] = numpy.concatenate(lst)
    return out, max_length


def list3d_to_array(lst, diodes):
    """ Convert list of ragged length 2-D arrays to 3-D numpy array.

    Parameters
    ----------
    lst : list of 2-D arrays
        List of arrays to normalize
    diodes : int
        Number of diodes in instrument array.

    Returns
    -------
    array
        3-d array of input arrays.
    """
    lengths = numpy.array([len(item) for item in lst])
    max_length = lengths.max()
    mask = lengths[:, None] > numpy.arange(max_length)
    out = numpy.zeros((mask.shape[0], mask.shape[1], diodes), dtype='uint8')
    out[mask] = numpy.concatenate(lst)
    return out


def convert_uint8_to_bitlist(image):
    """ Converts uint8 value to list of bits assuming little endianness.

    Parameters
    ----------
    image : int
        Int value to convert to binary.

    Returns
    -------
    array
        Array of bits corresponding to provided uint8 value.
    """
    img = numpy.array(image, dtype=numpy.uint8)
    bin_img = numpy.unpackbits(img, bitorder='little')
    return bin_img


def convert_uint32_to_bitlist(image):
    if type(image) is list:
        bin_img = []
        for im in image:
            bin_img.append(_convert_uint32_to_bitlist(im))
    else:
        bin_img = _convert_uint32_to_bitlist(image)
    return bin_img


def _convert_uint32_to_bitlist(image):
    bitstr = f'{image:032b}'
    bin_img = [int(b) for b in list(bitstr)]
    return bin_img


@njit
def _reshape_image(image, max_slices, diodes):
    """ Numba reshape image function for faster processing.
    """
    return numpy.reshape(image, (-1, max_slices // diodes, diodes))



#!/usr/bin/env python
# coding: utf-8

from concurrent.futures import FIRST_COMPLETED
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import wait
import csv
import datetime
import gc
import math
import os
import time
import uuid

import numpy
from tqdm import tqdm

from . import BinaryFile
from ..images import Images

MAX_PROCESSORS = 20


class SPECFile(BinaryFile):
    """ Class representing monoscale binary format for SPEC instruments.
    Implements methods specific to decompressing and processing images from
    this type of binary file.
    Class Attributes
    ----------------
    syncword : numpy array
        Numpy array of syncword used in DMT monoscale format.
    Parameters
    ----------
    filename : str
        Filename of current file.
    inst_name : str
        Name of current instrument.
    """
    syncword = numpy.array([3] * 128)

    def __init__(self, filename, inst_name, resolution):
        super().__init__(filename, inst_name, resolution)
        self.diodes = 128
        self.file_dtype = numpy.dtype([('year', 'u2'),
                                       ('month', 'u2'),
                                       ('weekday', 'u2'),
                                       ('day', 'u2'),
                                       ('hour', 'u2'),
                                       ('minute', 'u2'),
                                       ('second', 'u2'),
                                       ('ms', 'u2'),
                                       ('data', '(2048, )u2'),
                                       ('discard', 'u2')
                                       ])
        self.aux_channels = ['tas', 'clock_counts']

    def calc_buffer_datetimes(self):
        """ Calculates datetimes from bufffers read in from file and sets
        to datetimes class attribute.
        """
        self.datetimes = [datetime.datetime(d['year'],
                                            d['month'],
                                            d['day'],
                                            d['hour'],
                                            d['minute'],
                                            d['second'],
                                            d['ms'] * 1000) for d in self.data]
        self.datetimes = numpy.array(self.datetimes)

    def process_file(self, spiffile, processors=None, start=0, end=None):
        """ Method to process file using multiple processors. Calls
        process_image method implemented in children classes. Passes image
        data to spiffile object to be written to file each time 10,000 or more
        images are extracted from file.
        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        processors : int, optional
            Number of processors to use for parallel image processing. If none
            is provided, defaults to number of processors minus one.
        """

        if processors is None:
            processors = os.cpu_count()
            if os.cpu_count() > 1:
                processors -= 1

        if processors > MAX_PROCESSORS:
            processors = MAX_PROCESSORS

        spiffile.set_start_date(self.start_date.strftime('%Y-%m-%d %H:%M:%S %z'))

        if end is None:
            process_until = len(self.data)
        else:
            process_until = end - 1

        data_chunk = range(start, process_until)
        tot_h_images = 0
        tot_v_images = 0

        if self.name == '2DS':
            spiffile.create_inst_group(self.name + '-V')
            spiffile.create_inst_group(self.name + '-H')
            spiffile.set_filenames_attr(self.name + '-V', self.filename)
            spiffile.set_filenames_attr(self.name + '-H', self.filename)
            inst_groups = [self.name + '-H', self.name + '-V']
        elif self.name == 'HVPS4':
            spiffile.create_inst_group(self.name + '-V50', 50)
            spiffile.create_inst_group(self.name + '-V150', 150)
            spiffile.create_inst_group(self.name + '-H50', 50)
            spiffile.create_inst_group(self.name + '-H150', 150)
            spiffile.set_filenames_attr(self.name + '-V50', self.filename)
            spiffile.set_filenames_attr(self.name + '-V150', self.filename)
            spiffile.set_filenames_attr(self.name + '-H50', self.filename)
            spiffile.set_filenames_attr(self.name + '-H150', self.filename)
            inst_groups = [self.name + '-H50', self.name + '-V50',
                           self.name + '-H150', self.name + '-V150']
        else:
            spiffile.create_inst_group(self.name)
            spiffile.set_filenames_attr(self.name, self.filename)
            inst_groups = [self.name]

        spiffile.write_buffer_info(self.start_date, self.datetimes)

        if self.name == 'HVPS4':
            hk_file = str(self.filename) + 'HK'
        else:
            hk_file = None

        self.hk_data = FrameInfo(self.data, self.datetimes, hk_file)

        pbar1 = tqdm(desc='Processing frames',
                     total=process_until,
                     unit='frame')
        pbar2 = tqdm(desc='Writing frames',
                     total=process_until,
                     unit='frame')

        t00 = time.time()
        i = 0
        chunksize = 500
        max_write_queue = 8
        images_remaining = True
        futures = []
        # self.last_clocks = numpy.zeros(4)
        with ProcessPoolExecutor(max_workers=processors) as executor:
            while True:
                while len(futures) <= max_write_queue and images_remaining:
                    futures.append(executor.submit(self.process_frames,
                                                   data_chunk[i: i + chunksize]))
                    i += chunksize
                    if i >= process_until - start:
                        images_remaining = False
                    pbar1.update(chunksize)

                done, running = wait(futures, return_when=FIRST_COMPLETED)
                for f in done:
                    indx = futures.index(f)
                    if indx == 0:
                        pbar2.update(chunksize)
                        h_images, v_images, h150_images, v150_images, frames = f.result()
                        tot_h_images += len(h_images)
                        tot_v_images += len(v_images)
                        t0 = time.time()
                        self.partial_write(spiffile, h_images, v_images, h150_images, v150_images)
                        # print(f'Frame {frame} images '
                        #       f'written in {time.time() - t0:0.3f} seconds')
                        futures.pop(indx)
                        gc.collect()

                if not images_remaining and len(futures) == 0:
                    break

        pbar1.close()
        pbar2.close()
        print('Finished processing.')
        t0 = time.time()
        t11 = time.time()
        print(f'{tot_h_images}-H; {tot_v_images}-V images processed in {t11 - t00:0.3f} seconds')

        print('Recalculating image times...')

        self.calc_image_times(inst_groups, spiffile)

    def calc_image_times(self, img_groups, spiffile):
        times = numpy.array(self.datetimes, dtype='datetime64[ns]') - numpy.datetime64(self.start_date)
        secs = times.astype('timedelta64[s]')
        ns = times - secs
        datetimes = secs.astype(float) + ns.astype(float) * 1e-9

        for i, img_group in enumerate(img_groups):
            buffer_indx = spiffile.instgrps[img_group]['core']['buffer_index'][:]
            tas = spiffile.instgrps[img_group]['core']['tas'][:]
            counts = spiffile.instgrps[img_group]['core']['clock_counts'][:]

            buffer_time = datetimes[buffer_indx]

            delta_count = numpy.diff(counts)
            delta_buffer = numpy.diff(buffer_time)

            tas = tas[:-1]

            rollover = numpy.where((delta_count < 0) | (delta_buffer > 200))[0]

            delta_count[delta_count < 0] += 2 ** 32

            delta_time = delta_count * self.resolution / tas * 1e-6

            itimematch = numpy.where(delta_buffer > 0)[0] - 1

            newtime = buffer_time
            elapsed_time = numpy.zeros(len(newtime))

            if len(rollover) > 0:
                rollover = numpy.insert(rollover, -1, len(newtime))
            else:
                rollover = [len(newtime)]

            istart = 0
            for ro in rollover:
                istop = ro - 1
                elapsed_time[istart: istop] = numpy.cumsum(delta_time[istart: istop])

                matches = numpy.where((itimematch > istart) & (itimematch < istop))[0]
                offset = numpy.median(buffer_time[itimematch[matches]] - elapsed_time[itimematch[matches]])
                newtime[istart: istop] = elapsed_time[istart: istop] + offset
                istart = ro

            epoch_time = numpy.modf(newtime)
            secs = epoch_time[1]
            ns = epoch_time[0] * 1e9

            grp = spiffile.instgrps[img_group]['core']
            spiffile.write_variable(grp, 'image_sec', secs)
            spiffile.write_variable(grp, 'image_ns', ns)


    def partial_write(self, spiffile, h_p, v_p, h_p150=None, v_p150=None):
        """ Called each time number of unsaved processed images exceeds
        10,000. Writes H and V images to SPIF NetCDF file.
        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        h_p : Images object
            Images object containing processed H array image info.
        v_p : Images object
            Images object containing processed V array image info.
        h_p150 : Images object, optional
            Images object containing processed H array image info with 150µm resolution (in case of HVPS4)
        v_p150 : Images object, optional
            Images object containing processed V array image info with 150µm resolution (in case of HVPS4)
        """

        if self.name == '2DS':
            self._partial_write(spiffile, h_p, '-H')
            v_suffix = '-V'
        elif self.name == 'HVPS4':
            self._partial_write(spiffile, h_p, '-H50')
            self._partial_write(spiffile, h_p150, '-H150')
            self._partial_write(spiffile, v_p150, '-V150')
            v_suffix = '-V50'
        else:
            v_suffix = ''

        self._partial_write(spiffile, v_p, v_suffix)

        v_p.clear()
        h_p.clear()

    def _partial_write(self, spiffile, images, suffix=''):
        if len(images) > 0:
            images.conv_to_array(self.diodes)
            spiffile.write_images(self.name + suffix, images)

    def process_frames(self, frames):
        h_p = Images(self.aux_channels)
        v_p = Images(self.aux_channels)
        h150_p = Images(self.aux_channels)
        v150_p = Images(self.aux_channels)

        prev_counts = {'h': numpy.nan,
                       'v': numpy.nan,
                       'h150': numpy.nan,
                       'v150': numpy.nan}

        h_img = None
        v_img = None
        h_len = 0
        v_len = 0

        h150_img = None
        v150_img = None
        h150_len = 0
        v150_len = 0

        img_dict = {'v_len': v_len,
                    'v150_len': v150_len,
                    'v_img': v_img,
                    'v150_img': v150_img,
                    'h_len': h_len,
                    'h150_len': h150_len,
                    'h_img': h_img,
                    'h150_img': h150_img,
                    }

        for frame in frames:
            img_dict, prev_counts = self.process_frame(frame, img_dict, prev_counts)
            h_p.extend(img_dict['h_images'])
            v_p.extend(img_dict['v_images'])
            if img_dict['h150_images'] is not None:
                h150_p.extend(img_dict['h150_images'])
                v150_p.extend(img_dict['v150_images'])

        return h_p, v_p, h150_p, v150_p, frames

    def process_frame(self, frame, img_dict, prev_counts):
        """ Method to process single frame of image data. Decompresses image
        data from frame and passes resulting image data to process_image
        method to extract image info from image buffer.
        Parameters
        ----------
        frame : int
            Frame number in self.data to process.
        Returns
        -------
        Images object
            Images object containing H array image info in current frame.
        Images object
            Images object containing V array image info in current frame.
        """
        data = self.data[frame]
        record = data['data']

        # Define Images objects for current frame
        h_images = Images(self.aux_channels)
        v_images = Images(self.aux_channels)
        h150_images = Images(self.aux_channels)
        v150_images = Images(self.aux_channels)

        record_time = datetime.datetime(data['year'],
                                        data['month'],
                                        data['day'],
                                        data['hour'],
                                        data['minute'],
                                        data['second'],
                                        data['ms'] * 1000)

        # Set parameter defaults for current frame
        i = 0
        # h_img = None
        # v_img = None
        # h_len = 0
        # v_len = 0

        # h150_img = None
        # v150_img = None
        # h150_len = 0
        # v150_len = 0

        h_img = img_dict['h_img']
        v_img = img_dict['v_img']
        h_len = img_dict['h_len']
        v_len = img_dict['v_len']

        h150_img = img_dict['h150_img']
        v150_img = img_dict['v150_img']
        h150_len = img_dict['h150_len']
        v150_len = img_dict['v150_len']

        if self.name == 'HVPS4':
            hk_length = 83
        else:
            hk_length = 53
        tas = self.hk_data.tas[frame]

        # resolution = self.resolution * 1e-6  # 10µm in meters
        resolution = self.resolution

        while i < len(record) - 53:
            if tas > 0.1:
                tick_length = resolution / tas
            else:
                tick_length = 0

            if record[i] == 12883:  # equals '2S'
                h = self.decode_flags(record[i + 1])
                v = self.decode_flags(record[i + 2])
                image_count = record[i + 3]
                num_slices = record[i + 4]
                i += 5
                if (h['mismatch'] == 1) or (v['mismatch'] == 1):
                    i += h['n'] + v['n']
                else:
                    if h['n'] > 0:
                        h_decomp, h_count = self.process_image(record, i, h, True)

                        # Store dummy time for now since we will recompute
                        # following batch processing
                        image_time = record_time

                        if self.name == 'HVPS4' and h['fifo_array'] == 0:
                            # Calculate new image time due to slices being
                            # 3x larger than 50µm case
                            if h_count == 0:
                                h_count = prev_counts['h150']

                            h150_img, h150_len, h150_images = self.store_image(h,
                                                                               h150_img,
                                                                               h_decomp,
                                                                               h150_len,
                                                                               h150_images,
                                                                               image_time,
                                                                               frame,
                                                                               tas,
                                                                               h_count)
                            prev_counts['h150'] = h_count

                        else:
                            if h_count == 0:
                                h_count = prev_counts['h']
                            h_img, h_len, h_images = self.store_image(h,
                                                                      h_img,
                                                                      h_decomp,
                                                                      h_len,
                                                                      h_images,
                                                                      image_time,
                                                                      frame,
                                                                      tas,
                                                                      h_count)
                            prev_counts['h'] = h_count

                    if v['n'] > 0:
                        v_decomp, v_count = self.process_image(record, i, v)

                        # Store dummy time for now since we will recompute
                        # following batch processing
                        image_time = record_time

                        if self.name == 'HVPS4' and v['fifo_array'] == 0:
                            # Calculate new image time due to slices being
                            # 3x larger than 50µm case
                            if v_count == 0:
                                v_count = prev_counts['v150']
                            v150_img, v150_len, v150_images = self.store_image(v,
                                                                               v150_img,
                                                                               v_decomp,
                                                                               v150_len,
                                                                               v150_images,
                                                                               image_time,
                                                                               frame,
                                                                               tas,
                                                                               v_count)
                            prev_counts['v150'] = v_count

                        else:
                            if v_count == 0:
                                v_count = prev_counts['v']

                            v_img, v_len, v_images = self.store_image(v,
                                                                      v_img,
                                                                      v_decomp,
                                                                      v_len,
                                                                      v_images,
                                                                      image_time,
                                                                      frame,
                                                                      tas,
                                                                      v_count)
                            prev_counts['v'] = v_count

                    i += h['n'] + v['n']
            elif record[i] == 19787:  # equals 'MK'
                i += 23
            elif record[i] == 18507:  # equals 'HK'
                read_tas = numpy.array([record[i + 50], record[i + 49]],
                                       dtype='u2').view('float32')[0]
                tas_dec = read_tas % 1
                if read_tas < 1000 and read_tas > 0.1 and tas_dec == 0:
                    tas = read_tas
                i += hk_length
            elif record[i] == 20044:  # equals 'NL'
                break
            else:
                i += 1

        # images = [h_images, v_images, h150_images, v150_images]
        # for imgs in images:
        #     self.recalc_timestamps(record_time, imgs)

        # if frame % 1000 == 0:
        #     print('At frame ' + str(frame) + ' of ' + str(len(self.data)))

        # self.hk_data.set_prev_h_count(frame, prev_h_count)
        # self.last_v_timestamp[frame] = v_count

        img_dict = {'v_images': v_images,
                    'v150_images': v150_images,
                    'v_len': v_len,
                    'v150_len': v150_len,
                    'v_img': v_img,
                    'v150_img': v150_img,
                    'h_images': h_images,
                    'h150_images': h150_images,
                    'h_len': h_len,
                    'h150_len': h150_len,
                    'h_img': h_img,
                    'h150_img': h150_img,
                    }


        return img_dict, prev_counts

    def recalc_timestamps(self, record_time, imgs):
        tot_elapsed = numpy.nansum(imgs.dt)
        timestamp = []
        cum_time = 0
        for i, dt in enumerate(imgs.dt):
            try:
                delta_time = datetime.timedelta(microseconds=tot_elapsed - cum_time)
            except:
                print(tot_elapsed, cum_time, imgs.dt)
            time = record_time - delta_time
            epoch_time = math.modf((time - self.start_date).total_seconds())
            imgs.sec[i] = epoch_time[1]
            imgs.ns[i] = epoch_time[0] * 1e9
            cum_time += dt

    def decode_flags(self, record):
        """ Decode flags for given 16 bit record.
        Parameters
        ----------
        record : int
            16 bit record to extract flags from.
        Returns
        -------
        dict
            Dictionary object containing flags results for given record. Flags
            are:
                n - number of words, including timing words
                timing - 0 if timing words present, 1 if not found
                mismatch - timing word mismatch
                fifo_array - FIFO empty before timing word found (means image
                             was cut off); contains array flag for HVPS4
                overload - last two words of the record are overload timing words
        """
        flags = {}
        flags['n'] = record & 4095
        flags['timing'] = (record & (2 ** 12)) >> 12
        flags['mismatch'] = (record & (2 ** 13)) >> 13
        flags['fifo_array'] = (record & (2 ** 14)) >> 14
        flags['overload'] = (record & (2 ** 15)) >> 15

        return flags

    def process_image(self, record, i, p, h=False):
        """ Extracts image from record.
        Parameters
        ----------
        record : array
            Array of records in current buffer.
        i : int
            Current position in record array for processing.
        p : dict
            flags dict for current image
        Returns
        -------
        array
            Decompressed image.
        int
            Value of counter, if present. If not present, 0 is returned.
        """
        p_raw = record[i: i + p['n']]
        if p['timing'] == 0 and len(p_raw) > 2 and len(record) > i + p['n']:
            p_counter = (p_raw[-2] << 16) | p_raw[-1]
            # if h:
            #     print(p_raw[-2], p_raw[-1], p_counter, i, record[i-4])
            p_raw = p_raw[:-2]
        else:
            p_counter = 0

        if p['overload'] == 1:
            p_raw = p_raw[:-2]

        p_decomp = self.decompress_image(p_raw)

        return p_decomp, p_counter

    def store_image(self, p, p_img, p_decomp, p_len, images, p_time, frame, tas, clock_counts):
        """ If timeword present in current image, stores extracted image
        info in Images data object. Otherwise, concatenates current image
        data to existing image data.
        Parameters
        ----------
        p : dict
            flags dict for current image
        p_img : array
            Image array to write to file if timing word is present.
        p_decomp : array
            Current image array to append to p_img.
        p_len : int
            Length in slices of p_img.
        images : Images object
            Images object to save image data to.
        p_time : datetime object
            Time of current image.
        frame : int
            Buffer frame number currently being processed.
        tas : float
            True airspeed of current image.
        Returns
        -------
        array
            Image array of combined images (if no image written to Images
            object), or None if image was written.
        int
            Current length of image array.
        Images object
            Images object updated with new image information if timeword
            was present.
        """
        if p_img is None:
            p_img = p_decomp
            p_len = len(p_img) / 128
        else:
            p_img = numpy.concatenate((p_img, p_decomp), axis=0)
            p_len = len(p_img) / 128

        if p['timing'] == 0:
            if p_len > 0:
                epoch_time = math.modf((p_time - self.start_date).total_seconds())
                images.image.append(p_img)
                images.sec.append(epoch_time[1])
                images.ns.append(epoch_time[0] * 1e9)
                images.length.append(p_len)
                images.buffer_index.append(frame)
                images.tas.append(tas)
                images.clock_counts.append(clock_counts)
            p_img = None
            p_len = 0

        return p_img, p_len, images

    def decompress_image(self, img):
        """ Decompresses image image.
        Parameters
        ----------
        img : array
            Compressed image data.
        Returns
        -------
        array
            Array of decompressed image image.
        """
        img_decomp = []
        slice_decomp = []
        for line in img:
            if line == 32767:  # special case of 0x7fff
                img_decomp.extend([0] * 128)
            elif line == 16384:  # special case of 0x4000
                img_decomp.extend([1] * 128)
            else:
                timeslice = (line & (2 ** 15)) >> 15
                startslice = (line & (2 ** 14)) >> 14
                num_shaded = (line & 16256) >> 7
                num_clear = (line & 127)
                #print(line, timeslice, startslice, num_clear, num_shaded)
                if timeslice == 0:
                    if startslice == 1:
                        if len(slice_decomp) % 128 > 0:
                            slice_decomp.extend([0] * (128 - (len(slice_decomp) % 128)))
                        img_decomp.extend(slice_decomp)
                        slice_decomp = []

                    # if num_clear + num_shaded < 128:
                    slice_decomp.extend([0] * num_clear)
                    slice_decomp.extend([1] * num_shaded)

        img = numpy.logical_not(numpy.array(img_decomp))
        return img


class FrameInfo(object):

    def __init__(self, data, datetimes, hk_filename=None):

        n_frames = len(data)

        self.data = data
        self.datetimes = datetimes

        self.tas = numpy.ones(n_frames, dtype=float) * 0.1
        self.last_h_timestamp = numpy.ones(n_frames) * numpy.nan
        self.last_v_timestamp = numpy.ones(n_frames) * numpy.nan

        self.raw_counts = numpy.empty(n_frames, dtype=object)
        self.raw_counts[...] = [[] for i in range(n_frames)]

        if hk_filename is None:
            self.process_hk()
        else:
            self.process_hk_file(hk_filename)

    def process_hk(self):
        last_good_tas = numpy.nan
        for i, frame in enumerate(self.data):
            hk_indx = numpy.where(frame['data'] == 18507)[0]
            record = frame['data']
            tas_vals = []
            for indx in hk_indx:
                try:
                    tas = numpy.array([record[indx + 50],
                                       record[indx + 49]],
                                      dtype='u2').view('float32')[0]
                    tas_dec = tas % 1
                    if tas > 0.1 and tas < 1000 and tas_dec == 0:
                        tas_vals.append(tas)

                except IndexError:
                    pass

            if tas_vals:
                self.tas[i] = tas_vals[-1]
                last_good_tas = tas_vals[-1]
            else:
                self.tas[i] = last_good_tas

    def process_hk_file(self, filename):

        with open(filename) as fid:
            data = numpy.fromfile(fid, dtype='u2')

            possible_hk_index = numpy.where(data == 18507)[0]
            # print(len(possible_hk_index))

            tas = []
            hk_datetimes = []

            for indx in possible_hk_index:
                if data[indx - 8] > 2050 or data[indx - 8] < 1980:
                    continue
                record_time = datetime.datetime(data[indx - 8],
                                                data[indx - 7],
                                                data[indx - 5],
                                                data[indx - 4],
                                                data[indx - 3],
                                                data[indx - 2],
                                                data[indx - 1] * 1000)
                hk_datetimes.append(record_time)

                tas.append(numpy.array([data[indx + 76],
                                        data[indx + 75]],
                                       dtype='u2').view('float32')[0])

        datetimes = numpy.array(self.datetimes).astype('datetime64[us]')
        hk_datetimes = numpy.array(hk_datetimes).astype('datetime64[us]')
        for i, record_time in enumerate(datetimes):
            best_indx = numpy.where(hk_datetimes > record_time)[0]
            # print(best_indx[0], len(self.tas), len(tas), i, len(datetimes))
            self.tas[i] = tas[best_indx[0]]


class SpecHKFile(object):

    def __init__(self, filename):
        self.filename = filename
        self.process()

    def process(self):
        with open(self.filename) as fid:
            data = numpy.fromfile(fid, dtype='u2')

        possible_hk_index = numpy.where(data == 18507)[0]
        print(len(possible_hk_index))

        self.hk_timestamps = []
        self.tas = []

        for indx in possible_hk_index:
            record_time = datetime.datetime(data[indx - 8],
                                            data[indx - 7],
                                            data[indx - 5],
                                            data[indx - 4],
                                            data[indx - 3],
                                            data[indx - 2],
                                            data[indx - 1] * 1000)
            self.hk_timestamps.append(record_time)
            self.tas.append(numpy.array([data[indx + 76],
                                         data[indx + 75]],
                                        dtype='u2').view('float32')[0])

        self.timestamps = numpy.array(self.hk_timestamps)
        self.tas = numpy.array(self.tas)


class SpecHKData(object):

    def __init__(self, data):
        self.data = data
        self.process()

    def process(self):
        frame_tas_vals = []
        last_good_tas = numpy.nan
        for frame in self.data:
            hk_indx = numpy.where(frame['data'] == 18507)[0]
            record = frame['data']
            tas_vals = []
            for i, indx in enumerate(hk_indx):
                try:
                    tas = numpy.array([record[indx + 50],
                                       record[indx + 49]],
                                      dtype='u2').view('float32')[0]
                    tas_dec = tas % 1
                    if tas > 0.1 and tas < 1000 and tas_dec == 0:
                        tas_vals.append(tas)

                except IndexError:
                    pass

            if tas_vals:
                frame_tas_vals.append(tas_vals[-1])
                last_good_tas = tas_vals[-1]
            else:
                frame_tas_vals.append(last_good_tas)

        self.tas = numpy.array(frame_tas_vals)



#!/usr/bin/env python
# coding: utf-8

import configparser
import datetime
import os

import netCDF4 as nc
import numpy

from .utils import convert_datetimes_to_seconds

spif_version = '0.86'

TIME_CHUNK = 64 * 32
IMAGE_TIME_CHUNK = 64 * 8

class SPIFCore(object):
    """ Defines toplevel SPIF object, intended to pass information between
    binary file classes and SPIFFile class.

    Attributes
    ----------
    config : configparser object
        Attributes read from instrument ini for use in defining SPIF NetCDF
        file metadata.
    instfile : BinaryFile object
        Reference to BinaryFile type object representing binary file to be
        converted to SPIF format.
    spiffile : SPIFFile object
        Reference to SPIFFile object representing output file for converted
        binary data.

    Parameters
    ----------
    inst_class : BinaryFile class
        Specific child class of BinaryFile to use for reading and converting
        binary data.
    filename : str
        Path to binary file to convert to SPIF format.
    outfile : str
        Path of SPIF output file to create.
    config : str
        Path to config file to use for processing binary file.
    """

    def __init__(self, inst_class, filename, outfile, config, start_time = None, end_time = None):
        self.config = self.get_config(config)

        self.instfile = inst_class(filename,
                                   self.config['instrument']['instrument_name'],
                                   self.config['resolution']['value'])
        self.spiffile = SPIFFile(outfile,
                                 self.config)

        self.start_time = start_time
        self.end_time = end_time

    def process(self, processors=None):
        """ Launches class methods to 1) create SPIF output file; 2) read
        binary file blocks; 3) process binary file blocks and output to SPIF;
        and 4) close SPIF file.
        """
        self.spiffile.create_file()
        self.instfile.read()

        self.instfile.process_file(self.spiffile, processors)
        self.spiffile.close()

    def get_config(self, config_file):
        """ Given config filename, open and return configparser object.

        Parameters
        ----------
        config_file : str
            Path to config file. None can be passed if filename is given.

        Returns
        -------
        configparser object
            ConfigParser object containing the config parameters in ini file.

        Raises
        ------
        FileNotFoundError
            Error raised if:
                - given config file does not exist
                - no config file or filename passed in
                - default config file cant be found in path for given filename
        """
        config = configparser.ConfigParser(allow_no_value=True)
        config.optionxform = str

        if config_file is not None:
            if not os.path.exists(config_file):
                raise FileNotFoundError(
                    'Config file could not be found at indicated path.')
        else:
            if self.filename is None:
                raise FileNotFoundError('Config file has not been defined.')
            else:
                config_file = os.path.dirname(
                    self.filename) + os.path.sep + self.name + '.ini'
                if not os.path.exists(config_file):
                    raise FileNotFoundError(
                        'Config file could not be found at default file path.')

        config.read(config_file)

        return config


class SPIFFile(object):
    """ Class representing NetCDF file conforming to SPIF conventions.

    Class Attributes
    ----------
    image_ns_attrs : dict
        Dictionary containing standard attributes for image_ns variable.
    image_sec_attrs : dict
        Dictionary containing standard attributes for image_sec variable.
    image_image_attrs : dict
        Dictionary containing standard attributes for image_image variable.
    image_len_attrs : dict
        Dictionary containing standard attributes for image_len variable.
    buffer_sec_attrs : dict
        Dictionary containing standard attributes for buffer_sec variable.
    inst_scalars : dict
        Dictionary containing names and variable types for instrument-level
        scalar attributes.

    Attributes
    ----------
    outfile : str
        Name of output NetCDF file.
    attrs : configparser object
        ConfigParser object containing attributes read from ini file.
    root_attrs : configparser section
        'root' section taken from attrs attribute.
    instrgrps : dict
        Dictionary for each instrument in SPIFFile containing instrument name
        and reference to its group in the NetCDF file.

    Parameters
    ----------
    outfile : str
        Name of output NetCDF file.
    attrs : configparser object
        ConfigParser object containing attributes read from ini file.

    """

    image_ns_attrs = {'long_name': 'image arrival time in nanoseconds',
                      'units': 'ns since image_sec',
                      'ancillary_variables': 'image_sec'}

    image_sec_attrs = {'standard_name': 'time',
                       'long_name': 'image arrival time in seconds',
                       'timezone': 'UTC',
                       'units': 'seconds since start_date',
                       'strftime_format': '%F %T %z'}
    image_attrs = {'long_name': 'image array'}
    image_len_attrs = {'long_name': 'image length',
                       'units': 'number of slices'}

    buffer_sec_attrs = {'long_name': 'buffer time in seconds',
                        'timezone': 'UTC',
                        'units': 'seconds since start_date',
                        'strftime_format': '%F %T %z'}

    inst_scalars = {'pixels': 'i2',
                    'resolution': 'f4',
                    'resolution_err': 'f4',
                    'arm_separation': 'f4',
                    'arm_separation_err': 'f4',
                    'antishatter_tips': 'b',
                    'bpp': 'i2'}

    def __init__(self, outfile, attrs):

        self.outfile = outfile
        self.attrs = attrs
        self.root_attrs = attrs['root']
        self.root_attrs['conventions'] = f'SPIF-{spif_version}'
        self.root_attrs['history'] = datetime.datetime.now().isoformat()
        self.instgrps = {}

    def create_file(self, mode='w'):
        """ Creates SPIF NetCDF output file and sets root attributes.

        Parameters
        ----------
        mode : str, optional
            File creation mode for NetCDF file.
        """
        self.rootgrp = nc.Dataset(self.outfile, mode, format='NETCDF4')

        self.rootgrp.setncatts(self.root_attrs)

    def create_inst_group(self, inst_name, resolution=None):
        """ Given desired instrument name, creates instrument group and
        associated attributes and variables in NetCDF file.
        Parameters
        ----------
        inst_name : str
            Short name of instrument to use for group name in NetCDF file.
        resolution : int, optional
            Custom resolution to be used in case of HVPS4 where separate instrument
            groups are needed.
        """
        instgrp = self.rootgrp.createGroup(inst_name)
        self.instgrps[inst_name] = instgrp
        coregrp = instgrp.createGroup('core')
        auxgrp = instgrp.createGroup('aux')

        n_pixels = self.attrs.getint('pixels', 'value')

        instgrp.createDimension('Images', None)
        instgrp.createDimension('Buffers', None)
        instgrp.createDimension('Pixels', None)
        # instgrp.createDimension('Array', n_pixels)
        # instgrp.createDimension('Slices', None)

        instgrp.setncatts(self.attrs['instrument'])
        for var, var_type in self.inst_scalars.items():
            if self.attrs.has_section(var):
                value = self.attrs[var]['value']
                if var == 'antishatter_tips':
                    value = bool(value)
                if var == 'resolution' and resolution is not None:
                    value = resolution
                attrs = dict(self.attrs[var].items())
                attrs.pop('value')
                self.create_variable(instgrp,
                                     var,
                                     var_type,
                                     (),
                                     attrs,
                                     data=value)

        self.create_variable(coregrp,
                             'image_ns',
                             'i4',
                             ('Images',),
                             self.image_ns_attrs,
                             chunksizes=(TIME_CHUNK,))

        self.create_variable(coregrp,
                             'image_sec',
                             'i4',
                             ('Images',),
                             self.image_sec_attrs,
                             chunksizes=(TIME_CHUNK,))

        self.create_variable(coregrp,
                             'image',
                             'u1',
                             ('Pixels',),
                             self.image_attrs,
                             complevel=3,
                             chunksizes=(IMAGE_TIME_CHUNK * 8 * n_pixels,))

        self.create_variable(coregrp,
                             'image_len',
                             'u2',
                             ('Images'),
                             self.image_len_attrs,
                             chunksizes=(TIME_CHUNK,))

        self.create_variable(coregrp,
                             'buffer_index',
                             'u4',
                             ('Images'),
                             chunksizes=(TIME_CHUNK,))

        self.create_variable(coregrp,
                             'buffer_sec',
                             'i4',
                             ('Buffers',),
                             self.buffer_sec_attrs,
                             chunksizes=(TIME_CHUNK,))

    def set_start_date(self, start_date):
        """ Given reference time, sets start_date attribute in NetCDF file.

        Parameters
        ----------
        start_date : str
            Start date in YYYY-MM-DD HH:MM:SS Z of data being written to file.

        """
        self.rootgrp.setncattr('start_date', start_date)

    def set_filenames_attr(self, inst_name, filenames):
        if type(filenames) is list:
            filenames = ', '.join(os.path.abspath(filenames))
        else:
            filenames = os.path.abspath(filenames)

        self.instgrps[inst_name].setncattr('raw_filenames', filenames)

    def create_variable(self, dataset, name, dtype, dims, attrs=None,
                        data=None, zlib=True, complevel=3, chunksizes=None):
        """ Given a NetCDF dataset object, output name, data type, and
        NetCDF dimensions, create variable in given NetCDF dataset. If
        data object is passed, write the data out to the created variable
        in the NetCDF file.

        Parameters
        ----------
        dataset : NetCDF dataset or group object
            Dataset to write data to - can be root or group dataset in a NetCDF
            file.
        name : str
            Name to use for variable in NetCDF dataset.
        dtype : str
            Data type of data
        dims : tuple
            Tuple containing string names of NetCDF dimensions to use in for
            data.
        attrs : dict, optional
            Dictionary of attribute name/value pairs to assign to variable.
        data : scalar or array, optional
            Data to write to created variable in NetCDF file.
        zlib : bool, optional
            Flag to activate zlib compression on variable. Default is True.
        complevel : int, optional
            Compression level to use if zlib is true. Defaults to 3, max is 9.
        chunksizes : tuple, optional
            Tuple of chunksizes (one for each dimension) to use in creation
            of variable in NetCDF file.
        """

        if name == '':
            return

        fv = None
        if attrs is not None:
            if '_FillValue' in attrs:
                fv = numpy.array(attrs.pop('_FillValue'), dtype)

        i = 1
        basename = name
        while name in dataset.variables:
            name = basename + '_' + str(i)
            i += 1

        out = dataset.createVariable(name,
                                     dtype,
                                     dims,
                                     fill_value=fv,
                                     zlib=zlib,
                                     complevel=complevel,
                                     chunksizes=chunksizes)

        if attrs is not None:
            for key, val in attrs.items():
                if type(val) is bool:
                    attrs[key] = str(val)
            out.setncatts(attrs)

        if data is not None:
            out[:] = data

    def write_variable(self, dataset, name, data, start_val=None):
        """ Writes data to given variable in provided dataset.

        Parameters
        ----------
        dataset : NetCDF dataset or group object
            Dataset to write data to - can be root or group dataset in a NetCDF
            file.
        name : str
            Name of variable to write to.
        data : scalar or array
            Data to write to file
        start_val : int, optional
            Index of value to use for writing data, e.g. to append data to
            existing variable.
        """
        out = dataset[name]
        if start_val is not None:
            out[start_val:] = data
        else:
            out[:] = data

    def write_images(self, inst_name, images):
        """ Writes image information to NetCDF file. If image information
        already exists in NetCDF file, new image information is appended
        to the end of existing data.

        Parameters
        ----------
        inst_name : str
            Instrument name to write image information to.
        images : Images object
            Images object containing image information to write to file.
        """
        instgrp = self.instgrps[inst_name]
        coregrp = instgrp['core']
        dim_size = len(instgrp.dimensions['Images'])
        px_size = len(instgrp.dimensions['Pixels'])
        self.write_variable(coregrp, 'image_sec', images.sec, dim_size)
        self.write_variable(coregrp, 'image_ns', images.ns, dim_size)
        self.write_variable(coregrp, 'image_len', images.length, dim_size)
        self.write_variable(coregrp, 'buffer_index', images.buffer_index, dim_size)
        self.write_variable(coregrp, 'image', images.image, px_size)
        for key, val in images.__dict__.items():
            if key not in images.default_items:
                if key not in coregrp.variables.keys():
                    self.create_variable(coregrp,
                                         key,
                                         'f',
                                         ('Images',),
                                         attrs=None,
                                         data=val,
                                         zlib=True,
                                         chunksizes=(TIME_CHUNK,))
                else:
                    self.write_variable(coregrp, key, val, dim_size)

        self.rootgrp.sync()

    def write_images_with_extra_aux_dtypes(self, inst_name, images, aux_dtypes):
        """ Writes image information to NetCDF file. If image information
        already exists in NetCDF file, new image information is appended
        to the end of existing data.

        Parameters
        ----------
        inst_name : str
            Instrument name to write image information to.
        images : Images object
            Images object containing image information to write to file.
        """
        instgrp = self.instgrps[inst_name]
        coregrp = instgrp['core']
        dim_size = len(instgrp.dimensions['Images'])
        px_size = len(instgrp.dimensions['Pixels'])
        self.write_variable(coregrp, 'image_sec', images.sec, dim_size)
        self.write_variable(coregrp, 'image_ns', images.ns, dim_size)
        self.write_variable(coregrp, 'image_len', images.length, dim_size)
        self.write_variable(coregrp, 'buffer_index', images.buffer_index, dim_size)
        self.write_variable(coregrp, 'image', images.image, px_size)
        for key, val in images.__dict__.items():
            if key not in images.default_items:
                if key not in coregrp.variables.keys():
                    self.create_variable(coregrp,
                                         key,
                                         aux_dtypes[key],
                                         ('Images',),
                                         attrs=None,
                                         data=val,
                                         zlib=True,
                                         chunksizes=(TIME_CHUNK,))
                else:
                    self.write_variable(coregrp, key, val, dim_size)

        self.rootgrp.sync()

    # def write_aux(self, instgrp, aux_dim_name, aux_dim_length, aux_dict):

    #     if 'Aux' in instgrp.groups:
    #         auxgrp = instgrp['Aux']
    #     else:
    #         auxgrp = instgrp.createGroup('Aux')

    #     auxgrp.createDimension(aux_dim_name, (aux_dim_length,))

    #     for key, val in aux_dict.items():
    #         self.create_variable(auxgrp,
    #                              key,
    #                              'f',
    #                              (aux_dim_name,),
    #                              zlib=True
    #                              )

    def write_buffer_info(self, start_date, datetimes):
        """ Writes buffer header information to core group of NetCDF file.

        Parameters
        ----------
        start_date : datetime object
            Reference time to use for converting datetimes into seconds since
            start_date.
        datetimes : array of datetime objects
            Datetimes read from binary file buffer headers.
        """
        seconds = convert_datetimes_to_seconds(start_date, datetimes)
        for grp in self.instgrps.values():
            auxgrp = grp['core']
            self.write_variable(auxgrp, 'buffer_sec', seconds)

    def close(self):
        """ Closes NetCDF file.
        """
        self.rootgrp.close()



import numpy as np

housekeeping_template = np.dtype([
    ("housekeeping_packet_id", "i"), # This is to be filled just prior to writing into file
    ("buffer_id","i"),
    ("buffer_sec", "i"),
    ("buffer_ns", "i"),
    ("buffer_index","u2"),
    ("incomplete_packet", "u2"),
    ("hz_elem_0_voltage","f"),
    ("hz_elem_64_voltage","f"),
    ("hz_elem_127_voltage","f"),
    ("v_elem_0_voltage","f"),
    ("v_elem_64_voltage","f"),
    ("v_elem_127_voltage","f"),
    ("raw_pos_power_supply","f"),
    ("raw_neg_power_supply","f"),
    ("hz_arm_tx_temp","f"),
    ("hz_arm_rx_temp","f"),
    ("v_arm_tx_temp","f"),
    ("v_arm_rx_temp","f"),
    ("hz_tip_tx_temp","f"),
    ("hz_tip_rx_temp","f"),
    ("rear_opt_bridge_temp","f"),
    ("dsp_board_temp","f"),
    ("forward_vessel_temp","f"),
    ("hz_laser_temp","f"),
    ("v_laser_temp","f"),
    ("front_plate_temp","f"),
    ("power_supply_temp","f"),
    ("minus_5V_supply","f"),
    ("plus_5V_supply","f"),
    ("can_internal_pressure","f"),
    ("hz_elem_21_voltage","f"),
    ("hz_elem_42_voltage","f"),
    ("hz_elem_85_voltage","f"),
    ("hz_elem_106_voltage","f"),
    ("v_elem_21_voltage","f"),
    ("v_elem_42_voltage","f"),
    ("v_elem_85_voltage","f"),
    ("v_elem_106_voltage","f"),
    ("num_v_particles_detected","u2"),
    ("num_h_particles_detected","u2"),
    ("heater_outputs","u2"),
    ("h_laser_drive","f"),
    ("v_laser_drive","f"),
    ("hz_masked_bits","u2"),
    ("v_masked_bits","u2"),
    ("num_stereo_particles_detected","u2"),
    ("num_t_word_mismatch","u2"),
    ("num_slice_count_mismatch","u2"),
    ("num_hz_ov_periods","u2"),
    ("num_v_ov_periods","u2"),
    ("compression_conf","u2"),
    ("num_empty_fifo","u2"),
    ("spare_2","u2"),
    ("spare_3","u2"),
    ("tas","f"),
    ("timing_word_1","u2"),
    ("timing_word_2","u2")
])

def process_housekeeping(buffer_id, buffer_sec, buffer_ns, buffer_index, raw_housekeeping):
    housekeeping_container = np.zeros(1, dtype = housekeeping_template)

    housekeeping_container['buffer_id'][:] = buffer_id
    housekeeping_container['buffer_sec'][:] = buffer_sec
    housekeeping_container['buffer_ns'][:] = buffer_ns
    housekeeping_container['buffer_index'][:] = buffer_index

    if len(raw_housekeeping) < 53:
        raw_housekeeping = np.append( raw_housekeeping, [0] * (53 - len(raw_housekeeping)) )
        housekeeping_container['incomplete_packet'][:] = 1

    housekeeping_container["hz_elem_0_voltage"][:] =                 raw_housekeeping[1]  * 0.00244140625
    housekeeping_container["hz_elem_64_voltage"][:] =                raw_housekeeping[2]  * 0.00244140625
    housekeeping_container["hz_elem_127_voltage"][:] =               raw_housekeeping[3]  * 0.00244140625
    housekeeping_container["v_elem_0_voltage"][:] =                  raw_housekeeping[4]  * 0.00244140625
    housekeeping_container["v_elem_64_voltage"][:] =                 raw_housekeeping[5]  * 0.00244140625
    housekeeping_container["v_elem_127_voltage"][:] =                raw_housekeeping[6]  * 0.00244140625
    housekeeping_container["raw_pos_power_supply"][:] =              raw_housekeeping[7]  * 0.00488400488
    housekeeping_container["raw_neg_power_supply"][:] =              raw_housekeeping[8]  * 0.00488400488
    housekeeping_container["hz_arm_tx_temp"][:] =                    raw_housekeeping[9]  * 0.00244140625 + 1.6
    housekeeping_container["hz_arm_rx_temp"][:] =                    raw_housekeeping[10] * 0.00244140625 + 1.6
    housekeeping_container["v_arm_tx_temp"][:] =                     raw_housekeeping[11] * 0.00244140625 + 1.6
    housekeeping_container["v_arm_rx_temp"][:] =                     raw_housekeeping[12] * 0.00244140625 + 1.6
    housekeeping_container["hz_tip_tx_temp"][:] =                    raw_housekeeping[13] * 0.00244140625 + 1.6
    housekeeping_container["hz_tip_rx_temp"][:] =                    raw_housekeeping[14] * 0.00244140625 + 1.6
    housekeeping_container["rear_opt_bridge_temp"][:] =              raw_housekeeping[15] * 0.00244140625 + 1.6
    housekeeping_container["dsp_board_temp"][:] =                    raw_housekeeping[16] * 0.00244140625 + 1.6
    housekeeping_container["forward_vessel_temp"][:] =               raw_housekeeping[17] * 0.00244140625 + 1.6
    housekeeping_container["hz_laser_temp"][:] =                     raw_housekeeping[18] * 0.00244140625 + 1.6
    housekeeping_container["v_laser_temp"][:] =                      raw_housekeeping[19] * 0.00244140625 + 1.6
    housekeeping_container["front_plate_temp"][:] =                  raw_housekeeping[20] * 0.00244140625 + 1.6
    housekeeping_container["power_supply_temp"][:] =                 raw_housekeeping[21] * 0.00244140625 + 1.6
    housekeeping_container["minus_5V_supply"][:] =                   raw_housekeeping[22] * 0.00488400488
    housekeeping_container["plus_5V_supply"][:] =                    raw_housekeeping[23] * 0.00488400488
    housekeeping_container["can_internal_pressure"][:] =             raw_housekeeping[24] * 0.01835600000 - 3.846
    housekeeping_container["hz_elem_21_voltage"][:] =                raw_housekeeping[25] * 0.00244140625
    housekeeping_container["hz_elem_42_voltage"][:] =                raw_housekeeping[26] * 0.00244140625
    housekeeping_container["hz_elem_85_voltage"][:] =                raw_housekeeping[27] * 0.00244140625
    housekeeping_container["hz_elem_106_voltage"][:] =               raw_housekeeping[28] * 0.00244140625
    housekeeping_container["v_elem_21_voltage"][:] =                 raw_housekeeping[29] * 0.00244140625
    housekeeping_container["v_elem_42_voltage"][:] =                 raw_housekeeping[30] * 0.00244140625
    housekeeping_container["v_elem_85_voltage"][:] =                 raw_housekeeping[31] * 0.00244140625
    housekeeping_container["v_elem_106_voltage"][:] =                raw_housekeeping[32] * 0.00244140625
    housekeeping_container["num_v_particles_detected"][:] =          raw_housekeeping[33]
    housekeeping_container["num_h_particles_detected"][:] =          raw_housekeeping[34]
    housekeeping_container["heater_outputs"][:] =                    raw_housekeeping[35]
    housekeeping_container["h_laser_drive"][:] =                     raw_housekeeping[36] * 0.001220703
    housekeeping_container["v_laser_drive"][:] =                     raw_housekeeping[37] * 0.001220703
    housekeeping_container["hz_masked_bits"][:] =                    raw_housekeeping[38]
    housekeeping_container["v_masked_bits"][:] =                     raw_housekeeping[39]
    housekeeping_container["num_stereo_particles_detected"][:] =     raw_housekeeping[40]
    housekeeping_container["num_t_word_mismatch"][:] =               raw_housekeeping[41]
    housekeeping_container["num_slice_count_mismatch"][:] =          raw_housekeeping[42]
    housekeeping_container["num_hz_ov_periods"][:] =                 raw_housekeeping[43]
    housekeeping_container["num_v_ov_periods"][:] =                  raw_housekeeping[44]
    housekeeping_container["compression_conf"][:] =                  raw_housekeeping[45]
    housekeeping_container["num_empty_fifo"][:] =                    raw_housekeeping[46]
    housekeeping_container["tas"][:] =                               np.array(
                                                                        [raw_housekeeping[50],raw_housekeeping[49]],
                                                                        dtype=np.uint16
                                                                    ).view("f")[0]
    housekeeping_container["timing_word_1"][:] =                     raw_housekeeping[51]
    housekeeping_container["timing_word_2"][:] =                     raw_housekeeping[52]

    return housekeeping_container

# A utility method to create the housekeeping group in a netcdf file

def create_housekeeping_group(spiffile, inst_name):
    hk_group = spiffile.rootgrp[inst_name].createGroup('aux-housekeeping')
    hk_group.createDimension('Packet', None)

# A utility method to define the extra image related variables in a netcdf file

def add_housekeeping_variables(spiffile, inst_name, num_buffers):
    hk_group = spiffile.rootgrp[inst_name]['aux-housekeeping']

    if num_buffers < 128:
        chunk_size = num_buffers
    else:
        chunk_size = 128

    spiffile.create_variable( hk_group, "buffer_index",                 "i",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "buffer_sec",                   "i",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "buffer_ns",                    "i",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "incomplete_packet",            "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_0_voltage",            "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_64_voltage",           "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_127_voltage",          "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_0_voltage",             "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_64_voltage",            "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_127_voltage",           "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "raw_pos_power_supply",         "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "raw_neg_power_supply",         "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_arm_tx_temp",               "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_arm_rx_temp",               "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_arm_tx_temp",                "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_arm_rx_temp",                "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_tip_tx_temp",               "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_tip_rx_temp",               "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "rear_opt_bridge_temp",         "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "dsp_board_temp",               "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "forward_vessel_temp",          "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_laser_temp",                "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_laser_temp",                 "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "front_plate_temp",             "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "power_supply_temp",            "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "minus_5V_supply",              "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "plus_5V_supply",               "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "can_internal_pressure",        "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_21_voltage",           "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_42_voltage",           "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_85_voltage",           "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_elem_106_voltage",          "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_21_voltage",            "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_42_voltage",            "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_85_voltage",            "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_elem_106_voltage",           "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_v_particles_detected",     "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_h_particles_detected",     "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "heater_outputs",               "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "h_laser_drive",                "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_laser_drive",                "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "hz_masked_bits",               "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "v_masked_bits",                "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_stereo_particles_detected","u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_t_word_mismatch",          "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_slice_count_mismatch",     "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_hz_ov_periods",            "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_v_ov_periods",             "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "compression_conf",             "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "num_empty_fifo",               "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "spare_2",                      "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "spare_3",                      "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "tas",                          "f",  ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "timing_word_1",                "u2", ('Packet',), None, chunksizes=(chunk_size,) )
    spiffile.create_variable( hk_group, "timing_word_2",                "u2", ('Packet',), None, chunksizes=(chunk_size,) )



#!/usr/bin/env python
# coding: utf-8

import numpy

from spifpy.input import sea


def read_sea_image_data(filename, file_dtype, typ, inst_name=None):

    if typ == 'mono':
        probe_tag = sea.SEA_ACQUISITION_TYPE_CIP_IMAGE
    elif typ == 'grey':
        probe_tag = sea.SEA_ACQUISITION_TYPE_2D_GREY_ADVANCED
    elif typ == '2d':
        probe_tag = sea.SEA_ACQUISITION_TYPE_2D_MONO_IMAGE

    time, dset = read_sea_file(filename, probe_tag, inst_name)

    dset_len = len(dset)
    print(dset_len)

    if dset_len == 0:
        raise ValueError(f'No {inst_name} data present in {filename}.')

    year = [t.year for t in time]
    month = [t.month for t in time]
    day = [t.day for t in time]
    hour = [t.hour for t in time]
    minute = [t.minute for t in time]
    second = [t.second for t in time]
    ms = [int(t.microsecond / 1000) for t in time]

    data = numpy.empty(dset_len, dtype=file_dtype)

    data['year'] = year
    data['month'] = month
    data['day'] = day
    data['hour'] = hour
    data['minute'] = minute
    data['second'] = second
    data['ms'] = ms
    data['data'] = dset

    return data


def read_sea_tas(filename, typ, resolution):

    if typ is '2dc':
        tas_tag = 6  # 2D Mono TAS Factors

    time, dset = read_sea_file(filename, tas_tag)

    tas = []
    for data in dset:
        tas.append(calc_2dc_tas(data, resolution))

    return tas


def calc_2dc_tas(parameters, resolution):

    return 50000 * resolution * parameters[0] / parameters[1] / 1000000


def read_sea_file(filename, tag, inst_name=None):

    seafile = sea.SEAFile(filename)

    tags = seafile.get_tags_by_typ(tag)

    length = None

    if tag == sea.SEA_ACQUISITION_TYPE_CIP_IMAGE:
        length = 4096
    elif tag == sea.SEA_ACQUISITION_TYPE_2D_MONO_IMAGE:
        length = 1024

    if inst_name is not None:
        tags = [t for t in tags if inst_name.lower() in t.description.lower()]

    dset = []
    time = []
    for buf in seafile.iter_buffers({'tag_number': [t.tag_number
                                     for t in tags]}):

        time.append(sea.datetime_from_seatime(
            buf.get_dataset_by_tag_number(sea.TIME_TAG)))

        dset.append(buf.get_datasets_by_typ(tag)[0][:length])

    return time, dset



#!/usr/bin/env python
# coding: utf-8

import shutil
import argparse
import pathlib as pl

import pkg_resources

def copyconf():
    parser = argparse.ArgumentParser(description='Copies over a configuration file to current working directory')

    parser.add_argument('instrument',
                        type=str,
                        help= 'Name of the instrument to be processed')

    args = parser.parse_args()

    instrument = args.instrument

    instrument_config_name = f"{instrument}.ini"

    try:
        assert pkg_resources.resource_exists('spifpy', f"config/{instrument_config_name}")
    except AssertionError:
        print(f"ERROR : The supplied instrument name {instrument} cannot be found in the config list.")

    instrument_config_data = pkg_resources.resource_string('spifpy', f"config/{instrument_config_name}")
    aux_config_data = pkg_resources.resource_string('spifpy', f"config/aux_config.ini")

    dst_root = pl.Path.cwd()

    dst_instrument_config = dst_root / instrument_config_name
    dst_aux_config = dst_root / "aux_config.ini"

    dst_instrument_config.write_bytes(instrument_config_data)
    dst_aux_config.write_bytes(aux_config_data)



import numpy as np
from numba import njit

class ImageBufferDecompressor:

    def __init__(self) -> None:
        pass

    def decompress_buffer(self, buffer):
        decompressed_buffer = decompress_dmt_mono_buffer(buffer)

        return decompressed_buffer
@njit
def decompress_dmt_mono_buffer(data):

    decompressed_data = np.array([np.uint8(x) for x in range(0)], dtype = 'B')

    i = 0
    while i < len(data):
        rlehb_byte = data[i]

        all_zeros = rlehb_all_zeros(rlehb_byte)
        all_ones = rlehb_all_ones(rlehb_byte)
        all_dummy = rlehb_all_dummy(rlehb_byte)
        data_count = rlehb_count(rlehb_byte)

        i += 1

        if all_zeros:
            decompressed_data = np.append(decompressed_data, np.zeros(data_count, dtype = 'B'))
        elif all_ones:
            decompressed_data = np.append(decompressed_data, np.zeros(data_count, dtype = 'B') + np.uint8(255) )
        elif all_dummy:
            pass
        else:
            decompressed_data = np.append( decompressed_data, data[i:i + data_count])
            i += data_count

    return decompressed_data

"""
    Functions related to processing Run-Length Encoding Header Bytes(RHELBs)
"""

@njit
def rlehb_all_zeros(rlehb_byte):
    return (rlehb_byte & 0b10000000) >> 7

@njit
def rlehb_all_ones(rlehb_byte):
    return (rlehb_byte & 0b01000000) >> 6

@njit
def rlehb_all_dummy(rlehb_byte):
    return (rlehb_byte & 0b00100000) >> 5

@njit
def rlehb_count(rlehb_byte):
    return (rlehb_byte & 0b00011111) + 1



"""Probe types public imports."""

from .ProbeFactory import get_probe
from .probetypes import (
    CipProbe,
    CpiProbe,
    HvpsProbe,
    Hvps4V50Probe,
    Hvps4V150Probe,
    Hvps4H50Probe,
    Hvps4H150Probe,
    PipProbe,
    ProbeType,
    SpifProbe,
    TwoDcProbe,
    TwoDsHProbe,
    TwoDsVProbe,
)

__all__ = [
    "get_probe",
    "CipProbe",
    "CpiProbe",
    "HvpsProbe",
    "Hvps4V50Probe",
    "Hvps4V150Probe",
    "Hvps4H50Probe",
    "Hvps4H150Probe",
    "PipProbe",
    "ProbeType",
    "SpifProbe",
    "TwoDsHProbe",
    "TwoDsVProbe",
    "TwoDcProbe",
]



""" spifpy library public imports """

from pkg_resources import DistributionNotFound, get_distribution

from .core.dataset import SpifDataset

try:
    __version__ = get_distribution("spifpy").version
except DistributionNotFound:
    # package is not installed
    pass

__all__ = ["SpifDataset"]



from spifpy.input.binary_file import BinaryFile
from spifpy.input.dmt_binary_file import DMTBinaryFile
from spifpy.input.dmt_grey_file import DMTGreyFile
from spifpy.input.dmt_mono_file import DMTMonoFile
from spifpy.input.dmt_grey_mono_file import DMTGreyMonoFile
from spifpy.input.spec.spec_file import SPECFile
from spifpy.input.twod_file import TwoDFile

__all__ = [
    "BinaryFile",
    "DMTBinaryFile",
    "DMTGreyFile",
    "DMTGreyMonoFile",
    "DMTMonoFile",
    "SPECFile",
    "TwoDFile",
]



#!/usr/bin/env python
# coding: utf-8

import datetime
import math

import numpy

from spifpy.input.dmt_binary_file import DMTBinaryFile
from spifpy.images import Images
from spifpy.utils import bin_to_int, convert_uint8_to_bitlist


class DMTMonoFile(DMTBinaryFile):
    """ Class representing monoscale binary format for DMT instruments (CIP, PIP).
    Implements methods specific to decompressing and processing images from
    this type of binary file.

    Class Attributes
    ----------------
    syncword : numpy array
        Numpy array of syncword used in DMT monoscale format.

    Parameters
    ----------
    filename : str
        Filename of current file.
    inst_name : str
        Name of current instrument, e.g., CIP, PIP.
    """
    syncword = numpy.array([170] * 8, dtype='B')

    def __init__(self, filename, inst_name, resolution):
        super().__init__(filename, inst_name, resolution)
        self.diodes = 64
        self.file_dtype = numpy.dtype([('year', 'u2'),
                                       ('month', 'u2'),
                                       ('day', 'u2'),
                                       ('hour', 'u2'),
                                       ('minute', 'u2'),
                                       ('second', 'u2'),
                                       ('ms', 'u2'),
                                       ('weekday', 'u2'),
                                       ('data', '(4096, )B')
                                       ])

        self.aux_channels = ['image_count', 'dof_flag']

    def read(self):
        dummy = 0
        try:
            super().read()
        except ValueError:
            err = True
            while dummy < 32 and err:
                dummy += 1
                self.file_dtype = numpy.dtype([('year', 'u2'),
                                               ('month', 'u2'),
                                               ('day', 'u2'),
                                               ('hour', 'u2'),
                                               ('minute', 'u2'),
                                               ('second', 'u2'),
                                               ('ms', 'u2'),
                                               ('weekday', 'u2'),
                                               ('dummy', f'({dummy}, )B'),
                                               ('data', '(4096, )B')
                                               ])
                try:
                    super().read()
                    err = False
                except ValueError as v:
                    pass

    def process_frame(self, frame):
        """ Method to process single frame of image data. Decompresses image
        data from frame and passes resulting image data to extract_images
        method to extract image info from image buffer.

        Parameters
        ----------
        frame : int
            Frame number in self.data to process.

        Returns
        -------
        Images object
            Images object containing image info in current frame.
        """
        data = self.data[frame]
        record = data['data']
        i = 0
        frame_decomp = []
        while i < 4096:
            b1 = record[i]
            counts = (b1 & 31) + 1
            if b1 & 128 == 128:
                frame_decomp.extend([00] * counts)
            elif b1 & 64 == 64:
                frame_decomp.extend([255] * counts)
            elif b1 & 32 == 32:
                i = i + 1
                continue
            else:
                frame_decomp.extend(record[i + 1:i + counts + 1])
                i += counts
            i += 1
        date = datetime.datetime(data['year'],
                                 data['month'],
                                 data['day'])
        # if frame % 1000 == 0:
        #     print('At frame ' + str(frame) + ' of ' + str(len(self.data)))
        return self.extract_images(frame, frame_decomp, date)

    def extract_images(self, frame, frame_decomp, date):
        """ Extracts images and associated info from decompressed image
        buffer. The image image, length, timestamp in sec and ns, image
        count, dof flag, and buffer index are extracted for each image.

        Parameters
        ----------
        frame : int
            Frame number of current frame being processed.
        frame_decomp : array
            Array of decompressed image buffer.
        data : datetime object
            Date of current image buffer.

        Returns
        -------
        Images object
            Images object containing image data extracted from image
            buffer.
        """
        images = Images(self.aux_channels)

        ii = 0
        old_ii = 0
        while ii < len(frame_decomp) - 16:
            ii = self.search_syncword(frame_decomp, ii)
            ii_next = self.search_syncword(frame_decomp, ii + 8)
            # print(ii - old_ii)
            frameval = frame_decomp[ii: ii + 8]
            if ii >= len(frame_decomp) - 17:
                break
            header_bin = convert_uint8_to_bitlist(frame_decomp[ii + 8: ii + 16])

            image_count = bin_to_int(header_bin[0:16])
            hour = bin_to_int(header_bin[51:56])
            minute = bin_to_int(header_bin[45:51])
            second = bin_to_int(header_bin[39:45])
            millisecond = bin_to_int(header_bin[29:39])
            nanosecond_eigths = bin_to_int(header_bin[16:29])
            slice_count_old = bin_to_int(header_bin[57:64]) - 1
            slice_count = int((ii_next - ii) / 8) - 2
            # print(slice_count, slice_count_old)
            dof_flag = bin_to_int(header_bin[56:57])
            im = frame_decomp[ii + 16: ii + 16 + (slice_count * 8)]
            image = convert_uint8_to_bitlist(im)
            ii += 16 + (slice_count) * 8
            # old_ii = ii

            image_time = date + datetime.timedelta(hours=hour,
                                                   minutes=minute,
                                                   seconds=second)

            epoch_time = (image_time - self.start_date).total_seconds()
            images.ns.append(millisecond * 1e6 + nanosecond_eigths / 125)
            images.sec.append(epoch_time)
            if len(image) % self.diodes != 0:
                pad_amount = math.ceil(len(image) / self.diodes) * self.diodes - len(image)
                image = numpy.append(image, numpy.ones(pad_amount))
            images.length.append(len(image) // self.diodes)
            image = numpy.array(image)
            # image = numpy.reshape(image, (-1, self.diodes))
            images.image.append(image)
            images.image_count.append(image_count)
            images.buffer_index.append(frame)
            images.dof_flag.append(dof_flag)

        return images

    def search_syncword(self, frame_decomp, i):
        """ Searches image buffer for next image using syncword. If not found
        returns index of end of current image buffer.

        Parameters
        ----------
        frame_decomp : array
            Array of decompressed image buffer.
        i : int
            Index in frame_decomp to search from.

        Returns
        -------
        int
            Index value of start of next syncword.
        """
        try:
            while True:
                i = frame_decomp.index(self.syncword[0], i)
                if (frame_decomp[i: i + 8] == self.syncword).all():
                    return i
                i += 1
        except (ValueError, AttributeError, IndexError):
            return len(frame_decomp)



#!/usr/bin/env python
# coding: utf-8

import datetime

import numpy

from . import DMTBinaryFile
from ..images import Images
from ..utils import bin_to_int


class DMTGreyFile(DMTBinaryFile):
    """ Class representing greyscale binary format for DMT instruments (CIPGS).
    Implements methods specific to decompressing and processing images from
    this type of binary file.

    Class Attributes
    ----------------
    syncword : numpy array
        Numpy array of syncword used in DMT monoscale format.

    Parameters
    ----------
    filename : str
        Filename of current file.
    inst_name : str
        Name of current instrument, e.g., CIP, PIP.
    resolution : int
        Resolution in microns of probe
    """
    syncword = numpy.array([3] * 128)

    def __init__(self, filename, inst_name, resolution):
        super().__init__(filename, inst_name, resolution)
        self.diodes = 64
        self.file_dtype = numpy.dtype([('year', 'u2'),
                                       ('month', 'u2'),
                                       ('day', 'u2'),
                                       ('hour', 'u2'),
                                       ('minute', 'u2'),
                                       ('second', 'u2'),
                                       ('ms', 'u2'),
                                       ('weekday', 'u2'),
                                       ('data', '(4096, )B')
                                       ])
        self.aux_channels = ['tas', 'image_count']

    def process_frame(self, frame):
        """ Method to process single frame of image data. Decompresses image
        data from frame and passes resulting image data to extract_images
        method to extract image info from image buffer.

        Parameters
        ----------
        frame : int
            Frame number in self.data to process.

        Returns
        -------
        Images object
            Images object containing image info in current frame.
        """
        data = self.data[frame]
        record = data['data']
        i = 0
        frame_decomp = []
        pairs = [0, 0, 0, 0]
        while i < 4096:
            b1 = record[i]

            if b1 & 128 == 128:
                counts = b1 & 127
                frame_decomp.extend([pairs[3]] * counts)
            else:
                pairs = [int((b1 & 64) / 64),
                         int((b1 & 48) / 16),
                         int((b1 & 12) / 4),
                         int(b1 & 3)]
                if pairs[0] == 1:
                    frame_decomp.extend(pairs[1:])
                elif pairs[1] == 1:
                    frame_decomp.extend(pairs[2:])
                elif pairs[2] == 1:
                    frame_decomp.append(pairs[3])
            i += 1

        date = datetime.datetime(data['year'],
                                 data['month'],
                                 data['day'])
        # if frame % 1000 == 0:
        #     print('At frame ' + str(frame) + ' of ' + str(len(self.data)))

        return self.extract_images(frame, frame_decomp, date)

    def extract_images(self, frame, frame_decomp, date, mono=False):
        """ Extracts images and associated info from decompressed image
        buffer. The image image, length, timestamp in sec and ns, image
        count, dof flag, and buffer index are extracted for each image.

        Parameters
        ----------
        frame : int
            Frame number of current frame being processed.
        frame_decomp : array
            Array of decompressed image buffer.
        data : datetime object
            Date of current image buffer.

        Returns
        -------
        Images object
            Images object containing image data extracted from image
            buffer.
        """
        images = Images(self.aux_channels)

        ii = 0

        ii = self.search_syncword(frame_decomp, ii)
        while ii < len(frame_decomp) - 64:
            header = frame_decomp[ii: ii + 64]
            header8 = []
            for i in range(0, 64, 4):
                elem = 0
                for j in range(4):
                    h = header[i + j]
                    h = int((h & 2) / 2 + (h & 1) * 2)
                    elem += 2 ** (2 * (3 - j)) * h
                header8.append(elem)

            header8 = numpy.array(header8, dtype=numpy.uint8)
            header_bin = numpy.unpackbits(header8)

            slice_count = bin_to_int(header_bin[120:128])
            hour = bin_to_int(header_bin[115:120])
            minute = bin_to_int(header_bin[109:115])
            second = bin_to_int(header_bin[103:109])
            millisecond = bin_to_int(header_bin[93:103])
            microsecond = bin_to_int(header_bin[83:93])
            nanosecond_eigths = bin_to_int(header_bin[80:83])
            image_count = bin_to_int(header_bin[64:80])
            tas = bin_to_int(header_bin[56:64])
            image = frame_decomp[ii + 64: ii + 64 + (slice_count * 64)]
            if mono:
                for i, val in enumerate(image):
                    image[i] -= 2
            ii = ii + 64 + ((slice_count - 1) * 64)
            if ii < len(frame_decomp) - 128:
                ii = self.search_syncword(frame_decomp, ii)
            else:
                break

            image_time = date + datetime.timedelta(hours=int(hour),
                                                   minutes=int(minute),
                                                   seconds=int(second))

            epoch_time = (image_time - self.start_date).total_seconds()
            images.ns.append(millisecond * 1000 +
                             microsecond * 1e6 +
                             nanosecond_eigths / 125)
            images.sec.append(epoch_time)
            images.length.append(len(image) // self.diodes)
            image = numpy.array(image)
            # image = numpy.reshape(image, (-1, self.diodes))
            images.image.append(image)
            images.buffer_index.append(frame)
            images.tas.append(tas)
            images.image_count.append(image_count)

        return images

    def search_syncword(self, frame_decomp, i):
        """ Searches image buffer for next image using syncword. If not found
        returns index of end of current image buffer.

        Parameters
        ----------
        frame_decomp : array
            Array of decompressed image buffer.
        i : int
            Index in frame_decomp to search from.

        Returns
        -------
        int
            Index value of start of next syncword.
        """
        try:
            while True:
                i = frame_decomp.index(self.syncword[0], i)
                if (frame_decomp[i: i + 128] == self.syncword).all():
                    break
                i += 1

            while frame_decomp[i] == 3:
                i += 1
            return i
        except (ValueError, AttributeError, IndexError):
            return len(frame_decomp)



#!/usr/bin/env python
# coding: utf-8

from . import DMTGreyFile


class DMTGreyMonoFile(DMTGreyFile):

    def extract_images(self, frame, frame_decomp, date):
        images = super().extract_images(frame, frame_decomp, date, mono=True)

        return images



import numpy as np

from numba import njit
from math import ceil

from spifpy.spif import TIME_CHUNK

class ImageMetadataContainer:

    def __init__(self) -> None:
        self.hours = 0
        self.minutes = 0
        self.seconds = 0
        self.milliseconds = 0
        self.nanoseconds = 0

        self.particle_count = 0
        self.slice_count = 0
        self.dof_flag = 0

class ImageMetadataProcessor:

    def __init__(self) -> None:
        pass

    def process_image_metadata(self, image):
        timestamp_bitlist = get_timestamp_bitlist(image)

        metadata_container = ImageMetadataContainer()

        metadata_container.hours        = get_hours(timestamp_bitlist)
        metadata_container.minutes      = get_minutes(timestamp_bitlist)
        metadata_container.seconds      = get_seconds(timestamp_bitlist)
        metadata_container.milliseconds = get_milliseconds(timestamp_bitlist)
        metadata_container.nanoseconds  = get_nanoseconds(timestamp_bitlist)

        metadata_container.particle_count = get_particle_count(image)
        metadata_container.slice_count    = get_slice_count(image)
        metadata_container.dof_flag       = get_dof(image)

        return metadata_container

def get_timestamp_bitlist(image):
    timestamp_bytes = image[6:1:-1]

    timestamp_bitlist = int_to_bin(timestamp_bytes, big_endian = True)

    return timestamp_bitlist

def get_hours(timestamp_bitlist):
    hours = bin_to_int(timestamp_bitlist[0:5], big_endian = True)
    return hours

def get_minutes(timestamp_bitlist):
    minutes = bin_to_int(timestamp_bitlist[5:11], big_endian = True)
    return minutes

def get_seconds(timestamp_bitlist):
    seconds = bin_to_int(timestamp_bitlist[11:17], big_endian = True)
    return seconds

def get_milliseconds(timestamp_bitlist):
    milliseconds = bin_to_int(timestamp_bitlist[17:27], big_endian = True)
    return milliseconds

def get_nanoseconds(timestamp_bitlist):
    nanoseconds = bin_to_int(timestamp_bitlist[27:40], big_endian = True) * 125
    return nanoseconds

@njit
def get_particle_count(image):
    particle_count = image[:2].view(np.dtype('<u2'))[0]
    return particle_count

@njit
def get_slice_count(image_data):
    return(image_data[7] & 0b11111110) >> 1

@njit
def get_dof(image_data):
    return(image_data[7] & 0b00000001)


class ImageDataContainer:

    def __init__(self) -> None:
        self.image = np.array([], dtype = np.dtype('B'))
        self.image_len = 0

class ImageDataProcessor:

    def __init__(self) -> None:
        pass

    def process_image_data(self, decompressed_image):


        image_container = ImageDataContainer()

        image = decompressed_image[8:] # skipping the header

        image_len = len(image)

        if image_len % 8 != 0:
            padded_image_len = int( ceil(image_len/8)*8 )
            padding = np.zeros(padded_image_len - image_len, dtype = np.dtype('B'))
            image = np.append(image, padding)

        image = np.reshape(image, (-1, 8))
        image = np.fliplr(image)
        image = np.ravel(image)

        image_bitlist = int_to_bin(image)

        image_container.image = image_bitlist
        image_container.image_len = int(len(image_bitlist) / 64)

        return image_container

class AssembledImageContainer:

    def __init__(self) -> None:
        self.image_sec = np.array([], dtype = np.dtype('u4'))
        self.image_ns = np.array([], dtype = np.dtype('u4'))
        self.buffer_index = np.array([], dtype = np.dtype('u4'))
        self.buffer_sec = np.array([], dtype = np.dtype('u4'))
        self.buffer_ns = np.array([], dtype = np.dtype('u4'))
        self.particle_count = np.array([], dtype = np.dtype('u2'))
        self.slice_count = np.array([], dtype = np.dtype('B'))
        self.dof_flag = np.array([], dtype = np.dtype('B'))
        self.image_len = np.array([], dtype = np.dtype('u2'))
        self.image = np.array([], dtype = np.dtype('B'))

class ImageAssembler:

    def __init__(self) -> None:
        pass

    def assemble_images(self, metadata_containers, image_containers):
        assembled_image_container = AssembledImageContainer()

        assembled_image_container = self.set_buffer_info(
            assembled_image_container,
            num_images = len(metadata_containers)
        )

        assembled_image_container = self.set_image_sec(assembled_image_container, metadata_containers)
        assembled_image_container = self.set_image_ns(assembled_image_container, metadata_containers)
        assembled_image_container = self.set_particle_count(assembled_image_container, metadata_containers)
        assembled_image_container = self.set_slice_count(assembled_image_container, metadata_containers)
        assembled_image_container = self.set_dof_flag(assembled_image_container, metadata_containers)

        assembled_image_container = self.set_image_len(assembled_image_container, image_containers)
        assembled_image_container = self.set_image(assembled_image_container, image_containers)

        return assembled_image_container

    def set_buffer_info(self, assembled_record_container, num_images):
        assembled_record_container.buffer_index = np.zeros(num_images, dtype = np.dtype('u4'))
        assembled_record_container.buffer_sec = np.zeros(num_images, dtype = np.dtype('u4'))
        assembled_record_container.buffer_ns = np.zeros(num_images, dtype = np.dtype('u4'))

        return assembled_record_container

    def set_image_sec(self, assembled_record_container, metadata_containers):

        assembled_record_container.image_sec = np.zeros(len(metadata_containers), dtype=np.dtype('u4'))

        for i in range(len(metadata_containers)):

            assembled_record_container.image_sec[i] = \
              metadata_containers[i].hours*3600.0 \
            + metadata_containers[i].minutes*60 \
            + metadata_containers[i].seconds

        return assembled_record_container

    def set_image_ns(self, assembled_record_container, metadata_containers):

        assembled_record_container.image_ns = np.zeros(len(metadata_containers), dtype=np.dtype('u4'))

        for i in range(len(metadata_containers)):

            assembled_record_container.image_ns[i] = \
              metadata_containers[i].milliseconds*1e6 \
            + metadata_containers[i].nanoseconds

        return assembled_record_container

    def set_particle_count(self, assembled_record_container, metadata_containers):

        assembled_record_container.particle_count = np.zeros(len(metadata_containers), dtype=np.dtype('u2'))

        for i in range(len(metadata_containers)):
            assembled_record_container.particle_count[i] = metadata_containers[i].particle_count

        return assembled_record_container

    def set_slice_count(self, assembled_record_container, metadata_containers):

        assembled_record_container.slice_count = np.zeros(len(metadata_containers), dtype=np.dtype('B'))

        for i in range(len(metadata_containers)):
            assembled_record_container.slice_count[i] = metadata_containers[i].slice_count

        return assembled_record_container

    def set_dof_flag(self, assembled_record_container, metadata_containers):

        assembled_record_container.dof_flag = np.zeros(len(metadata_containers), dtype=np.dtype('B'))

        for i in range(len(metadata_containers)):
            assembled_record_container.dof_flag[i] = metadata_containers[i].dof_flag

        return assembled_record_container

    def set_image_len(self, assembled_record_container, image_containers):

        assembled_record_container.image_len = np.zeros(len(image_containers), dtype=np.dtype('u2'))

        for i in range(len(image_containers)):
            assembled_record_container.image_len[i] = image_containers[i].image_len

        return assembled_record_container

    def set_image(self, assembled_record_container, image_containers):

        assembled_record_container.image = np.zeros([], dtype=np.dtype('B'))

        image_arr = [x.image for x in image_containers]

        assembled_record_container.image = np.concatenate(image_arr)

        return assembled_record_container

def get_image_boundaries(data, syncword_sequence_inds):
    image_bound_start = syncword_sequence_inds
    image_bound_end = np.zeros_like(syncword_sequence_inds)
    image_bound_end[0:-1] = syncword_sequence_inds[1:]
    image_bound_end[-1] = len(data)

    image_bounds = np.vstack([image_bound_start, image_bound_end])
    image_bounds = image_bounds.T

    return image_bounds

def find_syncword_sequences(data):

    syncword = 0xAA
    conv_kernel = np.array([1,1,1,1,1,1,1,1], dtype = 'B')
    syncword_inds = (data == syncword)

    """
    'valid' makes the convolution start inside the array like this:
    |0,1,2|
    |0,1,2,3,4,5....|

    instead of this

    |0,1,2|
          |0,1,2,3,4,5...|
    """

    syncword_inds_conv = np.convolve(syncword_inds, conv_kernel, 'valid')

    syncword_sequence_inds = np.argwhere(syncword_inds_conv == 8)
    syncword_sequence_inds = np.ravel(syncword_sequence_inds)

    return syncword_sequence_inds

@njit
def int_to_bin(arr, big_endian = True, flat = True):
    val_bitstream = np.zeros((len(arr),8), dtype = np.dtype('B'))

    for i in range(len(arr)):
        for j in range(8):
            val_bitstream[i,j] = (arr[i] >> j) & 0b00000001

    if big_endian:
        val_bitstream = np.fliplr(val_bitstream)

    val_bitstream_flat = np.ravel(val_bitstream)

    return val_bitstream_flat

@njit
def bin_to_int(bin_list, big_endian=True):

    if big_endian:
        step = 1
    else:
        step = -1

    bin_list = bin_list[::step]
    bin_int = 0
    for bit in bin_list:
        bin_int = (bin_int << 1) | bit
    return int(bin_int)

def add_auxiliary_core_variables(spiffile, inst_name):
        coregrp = spiffile.rootgrp[inst_name]['core']

        spiffile.create_variable(
            coregrp,
            'dof_flag',
            'B',
            ('Images',),
            {
                'long_name':'Depth of Field Flag',
                'units':'boolean'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'particle_count',
            'u2',
            ('Images',),
            {
                'long_name':'Total counts of all particles',
                'units':'Counts'
            },
            chunksizes=(TIME_CHUNK,)
        )

        spiffile.create_variable(
            coregrp,
            'slice_count',
            'B',
            ('Images',),
            {
                'long_name':'Number of slices for this image',
                'units':'None'
            },
            chunksizes=(TIME_CHUNK,)
        )



import numpy as np

from spifpy.input.spec.image import ImageMetadataProcessor
from spifpy.input.spec.image import RawImageExtractor
from spifpy.input.spec.image import RawImageDecoder
from spifpy.input.spec.image import DecodedImageDecompressor
from spifpy.input.spec.image import ImageRecordAssembler

from spifpy.input.spec import housekeeping

DATA_FLAG = (ord('2') << 8) + ord('S')
HK_FLAG = (ord('H') << 8) + ord('K')
MK_FLAG = (ord('M') << 8) + ord('K')

mask_template = np.dtype([
    ("flag","u2"),
    ("timing_word", "int32"),
    ("hz_mask_byte_1", "u2"),
    ("hz_mask_byte_2", "u2"),
    ("hz_mask_byte_3", "u2"),
    ("hz_mask_byte_4", "u2"),
    ("hz_mask_byte_5", "u2"),
    ("hz_mask_byte_6", "u2"),
    ("hz_mask_byte_7", "u2"),
    ("hz_mask_byte_8", "u2"),
    ("v_mask_byte_1", "u2"),
    ("v_mask_byte_2", "u2"),
    ("v_mask_byte_3", "u2"),
    ("v_mask_byte_4", "u2"),
    ("v_mask_byte_5", "u2"),
    ("v_mask_byte_6", "u2"),
    ("v_mask_byte_7", "u2"),
    ("v_mask_byte_8", "u2"),
    ("timing_word_start", "int32"),
    ("timing_word_end", "int32")
])

class Buffer:

    def __init__(self, buffer_id, buffer_ts, buffer):
        self.buffer_id = buffer_id
        self.buffer_sec = self.get_buffer_sec(buffer_ts)
        self.buffer_ns = self.get_buffer_ns(buffer_ts)
        self.buffer = buffer

        self.metadata_processor = ImageMetadataProcessor()
        self.raw_image_extractor = RawImageExtractor()
        self.raw_image_decoder = RawImageDecoder()
        self.decoded_image_decompressor = DecodedImageDecompressor()
        self.image_record_assembler = ImageRecordAssembler()

        self.metadata_containers = []
        self.raw_image_containers = []
        self.image_timewords = []
        self.decoded_image_containers = []

        self.assembled_images = {
            'h':None,
            'v':None
        }

        self.housekeeping = np.array([], dtype = housekeeping.housekeeping_template)
        self.masks = []

        self.alloc_frames()

        self.extract_raw_images()
        self.decode_images()
        self.decompress_images()

        self.assemble_image_records()

    def get_buffer_sec(self, buffer_ts):
        return int(np.floor(buffer_ts))

    def get_buffer_ns(self, buffer_ts):
        return int( (buffer_ts - np.floor(buffer_ts))*1e9 )

    def alloc_frames(self):

        buffer_index = 0

        while buffer_index <  2048 - 5:
                flag = self.buffer[buffer_index]

                if flag == DATA_FLAG:
                    buffer_index = self.get_image_metadata_container(buffer_index)
                elif flag == HK_FLAG:
                    self.process_housekeeping(buffer_index)
                    buffer_index += 53
                elif flag == MK_FLAG:
                    self.process_masks(buffer_index)
                    buffer_index += 23
                else:
                    buffer_index += 1

    def get_image_metadata_container(self, buffer_index):
        image_metadata_container = self.metadata_processor.process_metadata(buffer_index, self.buffer)

        self.metadata_containers.append(image_metadata_container)

        return buffer_index + image_metadata_container.frame_len


    def extract_raw_images(self):

        raw_image_containers = [None]*len(self.metadata_containers)
        timeword_containers = [None]*len(self.metadata_containers)

        for i, metadata in enumerate(self.metadata_containers):
            raw_image_containers[i] = self.raw_image_extractor.extract_raw_images(
                metadata,
                self.buffer
            )

            timeword_containers[i] = self.raw_image_extractor.extract_image_timewords(
                metadata,
                self.buffer
            )

        self.raw_image_containers = raw_image_containers
        self.timeword_containers = timeword_containers

    def decode_images(self):
        decoded_image_containers = [None]*len(self.raw_image_containers)

        for i, raw_image_container in enumerate(self.raw_image_containers):
            decoded_image_containers[i] = self.raw_image_decoder.decode_dual_channel_images(
                raw_image_container
            )

        self.decoded_image_containers = decoded_image_containers

    def decompress_images(self):
        decompressed_image_containers = [None]*len(self.decoded_image_containers)

        for i, decoded_image_container in enumerate(self.decoded_image_containers):
            decompressed_image_containers[i] = self.decoded_image_decompressor.decompress_image(
                decoded_image_container
            )

        self.decompressed_image_containers = decompressed_image_containers

    def process_housekeeping(self, buffer_index):
        raw_housekeeping_packet = self.buffer[buffer_index:buffer_index + 53]

        self.housekeeping = np.append(
            self.housekeeping,
            housekeeping.process_housekeeping(
                self.buffer_id,
                self.buffer_sec,
                self.buffer_ns,
                buffer_index,
                raw_housekeeping_packet
            )
        )

    def process_masks(self, buffer_index):
        pass

    def assemble_image_records(self):

        assembled_images_h, assembled_images_v = self.image_record_assembler.assemble_images(
            self.buffer_id,
            self.buffer_sec,
            self.buffer_ns,
            self.metadata_containers,
            self.timeword_containers,
            self.decompressed_image_containers,
            self.housekeeping
        )

        self.assembled_images['h'] = assembled_images_h
        self.assembled_images['v'] = assembled_images_v



#!/usr/bin/env python
# coding: utf-8

from concurrent.futures import FIRST_COMPLETED
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import wait
import datetime
import os
import time

from tqdm import tqdm

from spifpy.input.binary_file import BinaryFile
from spifpy.input.input_utils import read_sea_image_data
from spifpy.images import Images

MAX_PROCESSORS = 20


class DMTBinaryFile(BinaryFile):
    """ Abstract class representing generic DMT binary file. Contains method
    used to process images using multiple processors.
    """

    def __init__(self, filename, inst_name, resolution):
        super().__init__(filename, inst_name, resolution)

    def read(self):
        file_ext = os.path.splitext(self.filename)[1]

        if file_ext == '.sea':
            self.read_sea()
        else:
            super().read()

    def read_sea(self, typ='mono'):

        self.data = read_sea_image_data(
            self.filename, self.file_dtype, typ, self.name)

        self.get_start_date()
        self.calc_buffer_datetimes()

    def process_file(self, spiffile, processors=None):
        """ Method to process file using multiple processors. Calls
        process_image method implemented in children classes. Passes image
        data to spiffile object to be written to file.

        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        processors : int, optional
            Number of processors to use for parallel image processing. If none
            is provided, defaults to number of processors minus one.
        """

        if processors is None:
            processors = os.cpu_count()
            if os.cpu_count() > 1:
                processors -= 1

        if processors > MAX_PROCESSORS:
            processors = MAX_PROCESSORS

        spiffile.set_start_date(self.start_date.strftime('%Y-%m-%d %H:%M:%S %z'))

        process_until = len(self.data)
        data_chunk = range(0, process_until)

        tot_images = 0

        spiffile.create_inst_group(self.name)
        spiffile.write_buffer_info(self.start_date, self.datetimes)
        spiffile.set_filenames_attr(self.name, self.filename)

        pbar1 = tqdm(desc='Processing frames',
                     total=len(self.data),
                     unit='frame')
        pbar2 = tqdm(desc='Writing frames',
                     total=process_until,
                     unit='frame')
        t00 = time.time()

        i = 0
        chunksize = 500
        max_write_queue = 8
        images_remaining = True
        futures = []
        with ProcessPoolExecutor(max_workers=processors) as executor:
            while True:
                while len(futures) <= max_write_queue and images_remaining:
                    futures.append(executor.submit(self.process_frames,
                                                   data_chunk[i: i + chunksize]))
                    i += chunksize
                    if i >= process_until:
                        images_remaining = False
                    pbar1.update(chunksize)

                done, running = wait(futures, return_when=FIRST_COMPLETED)
                for f in done:
                    indx = futures.index(f)
                    if indx == 0:
                        pbar2.update(chunksize)
                        images = f.result()
                        tot_images += len(images)
                        t0 = time.time()
                        images.conv_to_array(self.diodes)
                        spiffile.write_images(self.name, images)

                        futures.pop(indx)

                if not images_remaining and len(futures) == 0:
                    break

        pbar1.close()
        pbar2.close()
        t11 = time.time()
        print(f'{tot_images} images processed in {t11 - t00:0.3f} seconds')

    def process_frames(self, frames):
        p = Images(self.aux_channels)

        for frame in frames:
            images = self.process_frame(frame)
            p.extend(images)

        return p

    def process_frame(self, frame):
        raise NotImplementedError

    def calc_buffer_datetimes(self):
        """ Calculates datetimes from bufffers read in from file and sets
        to datetimes class attribute.
        """
        self.datetimes = [datetime.datetime(d['year'],
                                            d['month'],
                                            d['day'],
                                            d['hour'],
                                            d['minute'],
                                            d['second'],
                                            d['ms'] * 1000) for d in self.data]



#!/usr/bin/env python
# coding: utf-8

import datetime
import math

import numpy as np

from tqdm import tqdm

from spifpy.input.binary_file import BinaryFile
from spifpy.input.dmt.mono.buffer import decompress_dmt_mono_buffer
from spifpy.input.dmt.mono.image import find_syncword_sequences, get_image_boundaries
from spifpy.input.dmt.mono.image import ImageMetadataProcessor, ImageDataProcessor, ImageAssembler, AssembledImageContainer
from spifpy.input.dmt.mono.image import add_auxiliary_core_variables

file_buffer_template = np.dtype([
    ('year', 'u2'),
    ('month', 'u2'),
    ('day', 'u2'),
    ('hour', 'u2'),
    ('minute', 'u2'),
    ('second', 'u2'),
    ('ms', 'u2'),
    ('weekday', 'u2'),
    ('data', '(4096, )B')
])

syncword = 0xAA

sync_sequence = np.array([syncword] * 8, dtype='B')

class DMTMonoFile(BinaryFile):
    """ Class representing monoscale binary format for DMT instruments (CIP, PIP).
    Implements methods specific to decompressing and processing images from
    this type of binary file.

    Class Attributes
    ----------------
    syncword : numpy array
        Numpy array of syncword used in DMT monoscale format.

    Parameters
    ----------
    filename : str
        Filename of current file.
    inst_name : str
        Name of current instrument, e.g., CIP, PIP.
    """

    def __init__(self, filename, inst_name, resolution):
        super().__init__(filename, inst_name, resolution)
        self.diodes = 64
        self.file_dtype = file_buffer_template

    def process_file(self, spiffile, processors = None):
        """
        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        """

        spiffile.set_start_date(self.start_date.strftime('%Y-%m-%d %H:%M:%S %z'))

        spiffile.create_inst_group(self.name)
        spiffile.write_buffer_info(self.start_date, self.datetimes)
        spiffile.set_filenames_attr(self.name, self.filename)

        add_auxiliary_core_variables(spiffile, self.name)

        data_processor = DMTMonoDataProcessor(self.data['data'])

        processed_data = data_processor.process_data()

        self.write_image_data(spiffile, self.name, processed_data)

    def write_image_data(self, spiffile, instrument, images):

        """ Called each time number of unsaved processed images exceeds
        10,000. Writes H and V images to SPIF NetCDF file.
        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        h_p : Images object
            Images object containing processed H array image info.
        v_p : Images object
            Images object containing processed V array image info.
        """

        instgrp = spiffile.rootgrp[instrument]
        coregrp = spiffile.rootgrp[instrument]['core']

        dim_size = len(instgrp.dimensions['Images'])
        px_size = len(instgrp.dimensions['Pixels'])

        # Primary, critical SPIF variables

        spiffile.write_variable(coregrp, 'image_sec', images.image_sec, dim_size)
        spiffile.write_variable(coregrp, 'image_ns', images.image_ns, dim_size)
        spiffile.write_variable(coregrp, 'image_len', images.image_len, dim_size)
        spiffile.write_variable(coregrp, 'buffer_index', images.buffer_index, dim_size)
        spiffile.write_variable(coregrp, 'image', images.image, px_size)

        # Auxiliary SPIF variables

        spiffile.write_variable(coregrp, 'dof_flag', images.dof_flag, dim_size)
        spiffile.write_variable(coregrp, 'particle_count', images.particle_count, dim_size)
        spiffile.write_variable(coregrp, 'slice_count', images.slice_count, dim_size)

        spiffile.rootgrp.sync()

class DMTMonoDataProcessor:

    def __init__(self, data) -> None:
        self.data = data
        self.decompressed_data = np.array([], dtype='B')
        self.image_bounds = np.array([], dtype = 'i4')
        self.image_metadata_containers = []
        self.image_data = []

    def process_data(self):
        decompressed_data = self.decompress_data()
        image_bounds = self.get_image_bounds(decompressed_data)

        metadata_containers = self.process_image_metadata(decompressed_data, image_bounds)
        image_containers = self.process_image_data(decompressed_data, image_bounds)

        assembled_image_containers = self.assemble_image_data(metadata_containers, image_containers)

        return assembled_image_containers

    def decompress_data(self):
        pbar = tqdm(desc='Decompressing buffers',
             total=len(self.data),
             unit='buffers')


        decompressed_data = np.array([],dtype = 'B')

        for buffer in self.data:
            decompressed_buffer = decompress_dmt_mono_buffer(buffer)
            decompressed_data = np.append(decompressed_data, decompressed_buffer)
            pbar.update(1)

        pbar.close()

        return decompressed_data

    def get_image_bounds(self, decompressed_data):
        syncword_sequence_inds = find_syncword_sequences(decompressed_data)
        image_bounds = get_image_boundaries(decompressed_data, syncword_sequence_inds)

        return image_bounds

    def process_image_metadata(self, decompressed_data, image_bounds):
        pbar = tqdm(
            desc='Extracting image metadata',
            total=len(image_bounds),
            unit='images'
        )

        metadata_containers = [None]*len(image_bounds)

        metadata_processor = ImageMetadataProcessor()

        for i in range(len(image_bounds)):
            image = decompressed_data[image_bounds[i,0]:image_bounds[i,1]]
            image = image[8:] # cut out the string of 0xAA's that indicate the start of an image
            metadata_containers[i] = metadata_processor.process_image_metadata(image)
            pbar.update(1)

        pbar.close()

        return metadata_containers

    def process_image_data(self, decompressed_data, image_bounds):
        pbar = tqdm(
            desc='Extracting images',
            total=len(image_bounds),
            unit='images'
        )

        image_containers = [None]*len(image_bounds)

        image_processor = ImageDataProcessor()

        for i in range(len(image_bounds)):
            image = decompressed_data[image_bounds[i,0]:image_bounds[i,1]]
            image = image[8:] # cut out the string of 0xAA's that indicate the start of an image
            image_containers[i] = image_processor.process_image_data(image)
            pbar.update(1)

        pbar.close()

        return image_containers

    def assemble_image_data(self, metadata_containers, image_containers):

        assembled_image_container = AssembledImageContainer()

        image_assembler = ImageAssembler()

        assembled_image_container = image_assembler.assemble_images(metadata_containers, image_containers)

        return assembled_image_container



#!/usr/bin/env python
# coding: utf-8

import datetime

import numpy

class BinaryFile(object):
    """ Abstract class representing base BinaryFile. Should be subclassed by
    more specific binary file classes.

    Attributes
    ----------
    name : str
        Instrument name of current instrument.
    diodes : int
        Number of diodes in current instrument.
    aux_channels : list
        List of auxiliary channels to create in Particle object, e.g., tas,
        particle_number, etc.
    file_dtype : numpy dtype object
        Numpy dtype object defining binary file structure for specific instrument
        type.
    filename : str
        Filename of current file being referenced.
    data : numpy structured array
        Numpy structured array containing buffers read in from binary file.
    start_date : datetime object
        datetime object containing start date of data in current file.


    Parameters
    ----------
    filename : str
        Filename of current file.
    """

    def __init__(self, filename, inst_name, resolution):
        self.diodes = 0
        self.file_dtype = None
        self.aux_channels = None
        self.data = None
        self.start_date = None
        self.name = inst_name
        self.resolution = int(resolution)

        self.filename = filename

    def read(self):
        """ Reads binary data from file using file_dtype defined by subclass.
        """
        with open(self.filename) as fid:
            self.data = numpy.fromfile(fid, dtype=self.file_dtype)

        self.get_start_date()

        self.calc_buffer_datetimes()

    def get_start_date(self):
        """ Sets start_date attribute with first date found in buffer headers.
        """
        self.start_date = datetime.datetime(self.data['year'][0],
                                            self.data['month'][0],
                                            self.data['day'][0])

    def calc_buffer_datetimes(self):
        """ Calculates datetimes from buffers read in from file and sets
        to datetimes class attribute.

        This method is meant to be overriden, but has been reduced to
        calculating datetimes up to a second level of resolution. This is
        because the units past the level of one second might be different,
        (aka one probe could have it in ns, while another could have it
        in ms), so at least when you do call this function for any given
        probe, one can at least use a minimally informative list of datetimes.
        """

        self.datetimes = [datetime.datetime(d['year'],
                                            d['month'],
                                            d['day'],
                                            d['hour'],
                                            d['minute'],
                                            d['second']) for d in self.data]



class SEATag(object):
    def __init__(self, description, tag_number, samples, size,
            typ, parameter1, parameter2, parameter3):
        self.description = description
        self.tag_number = tag_number
        self.samples = samples
        self.size = size
        self.typ = typ
        self.parameter1 = parameter1
        self.parameter2 = parameter2
        self.parameter3 = parameter3

        for attr_name in vars(self):
            if attr_name != 'description':
                attr_val = getattr(self, attr_name)
                try:
                    setattr(self, attr_name, int(attr_val, 0))
                except ValueError as ve:
                    # if it didn't convert, it wasn't an int, so just leave it
                    pass
                except TypeError as te:
                    if te.args[0] == "int() can't convert non-string with explicit base":
                        # don't bother convertingthe non-string
                        pass
                    else:
                        raise

    def __repr__(self):
        parts = ['<sea.Tag ']
        for key, value in vars(self).items():
            parts.append('{:s}={}'.format(key,
                value if value is not None else 'undefined'))
        parts.append('>')
        return ' '.join(parts)

    # At the time of writing, and forseeable future, tag numbers can be
    # used as a primary identifier for tags within any given file from an
    # SEA M*00 system. (ie If two tags share the same tag number, all of
    # their remaining properties will be identical, and if two tags have
    # different tag numbers, some, but not necessarily all, of their
    # remaining properties will have different values). With this in mind,
    # __cmp__ and __hash__ have been implemented in the simplest way
    # possible to allow SEATag objects to be used as dictionary keys.
    def __hash__(self):
        return self.tag_number

    def __cmp__(self, other):
        return self.tag_number - other.tag_number


class M200Tag(SEATag):
    def __init__(self, description, buffer_number, tag_number, samples, size,
            typ, parameter1, parameter2, parameter3, address):
        super(M200Tag, self).__init__(description, tag_number, samples, size,
                typ, parameter1, parameter2, parameter3)

        self.buffer_number = buffer_number
        self.address = address

        for attr_name in ['buffer_number', 'address']:
            attr_val = getattr(self, attr_name)
            try:
                setattr(self, attr_name, int(attr_val, 0))
            except ValueError as ve:
                # if it didn't convert, it wasn't an int, so just leave it
                pass
            except TypeError as te:
                if te.message == \
                        "int() can't convert non-string with explicit base":
                    # don't bother converting the non-string
                    pass
                else:
                    raise

    def __repr__(self):
        parts = ['<m200.Tag ']
        for key, value in vars(self).items():
            parts.append('{:s}={}'.format(key,
                value if value is not None else 'undefined'))
        parts.append('>')
        return ' '.join(parts)


class M300Tag(SEATag):
    def __init__(self, description, tag_number, samples, state, size,
            typ, parameter1, parameter2, parameter3, board, sample_offset=0):
        super(M300Tag, self).__init__(description, tag_number, samples, size,
                typ, parameter1, parameter2, parameter3)

        self.state = state
        self.board = board
        self.sample_offset = sample_offset

        for attr_name in ['state', 'sample_offset']:
            attr_val = getattr(self, attr_name)
            try:
                setattr(self, attr_name, int(attr_val, 0))
            except ValueError as ve:
                # if it didn't convert, it wasn't an int, so just leave it
                pass
            except TypeError as te:
                if te.args[0] == "int() can't convert non-string with explicit base":
                    # don't bother convertingthe non-string
                    pass
                else:
                    raise

    def __repr__(self):
        parts = ['<m300.Tag ']
        for key, value in vars(self).items():
            parts.append('{:s}={}'.format(key,
                value if value is not None else 'undefined'))
        parts.append('>')
        return ' '.join(parts)



import datetime

import numpy as np
from scipy.interpolate import interp1d

from spifpy.input.binary_file import BinaryFile
from spifpy.input.spec.buffer import Buffer
from spifpy.input.spec.image import AssembledImageRecordContainer
from spifpy.input.spec.housekeeping import add_housekeeping_variables, housekeeping_template, create_housekeeping_group

from spifpy.input.spec.image import add_auxiliary_core_variables

from tqdm import tqdm

import joblib as jb

file_buffer_template = np.dtype([
    ('year', 'u2'),
    ('month', 'u2'),
    ('weekday', 'u2'),
    ('day', 'u2'),
    ('hour', 'u2'),
    ('minute', 'u2'),
    ('second', 'u2'),
    ('ms', 'u2'),
    ('data', '(2048, )u2'),
    ('discard', 'u2')
])

class SPECFile(BinaryFile):

    """ Class representing monoscale binary format for SPEC instruments.
    Implements methods specific to decompressing and processing images from
    this type of binary file.
    Parameters
    ----------
    filename : str
        Filename of current file.
    inst_name : str
        Name of current instrument.
    """

    def __init__(self, filename, inst_name, resolution):

        super().__init__(filename, inst_name, resolution)
        self.diodes = 128
        self.file_dtype = file_buffer_template
        self.aux_channels = None
        self.chunk_size = 25


    def calc_buffer_datetimes(self):

        """ Calculates datetimes from buffers read in from file and sets
        to datetimes class attribute.
        """

        self.datetimes = [datetime.datetime(d['year'],
                                            d['month'],
                                            d['day'],
                                            d['hour'],
                                            d['minute'],
                                            d['second'],
                                            d['ms'] * 1000) for d in self.data]

        self.timestamps = [(x - self.start_date).total_seconds() for x in self.datetimes]

    def process_file(self, spiffile, processors):

        spiffile.set_start_date(self.start_date.strftime('%Y-%m-%d %H:%M:%S %z'))

        group_no_channel = self.name
        group_h_channel = self.name + '-H'
        group_v_channel = self.name + '-V'

        if self.name == '2DS':
            spiffile.create_inst_group(group_h_channel)
            spiffile.create_inst_group(group_v_channel)
            spiffile.set_filenames_attr(group_h_channel, self.filename)
            spiffile.set_filenames_attr(group_v_channel, self.filename)

            add_auxiliary_core_variables(spiffile, group_h_channel)
            add_auxiliary_core_variables(spiffile, group_v_channel)

            create_housekeeping_group(spiffile, group_h_channel)
            create_housekeeping_group(spiffile, group_v_channel)

            add_housekeeping_variables(spiffile, group_h_channel, len(self.data))
            add_housekeeping_variables(spiffile, group_v_channel, len(self.data))

        else:
            spiffile.create_inst_group(group_no_channel)
            spiffile.set_filenames_attr(group_no_channel, self.filename)

            create_housekeeping_group(spiffile, group_no_channel)

            add_auxiliary_core_variables(spiffile, group_no_channel)
            add_housekeeping_variables(spiffile, group_no_channel, len(self.data))


        spiffile.write_buffer_info(self.start_date, self.datetimes)

        pbar1 = tqdm(desc='Processing frames',
                     total=len(self.data),
                     unit='frame')

        chunk_intervals = self.get_chunk_intervals()

        for i in range(len(chunk_intervals) - 1):
            chunk_slice = slice(chunk_intervals[i], chunk_intervals[i + 1])

            buffer_ids = [x for x in range(chunk_intervals[i], chunk_intervals[i + 1])]
            buffer_timestamps = self.timestamps[chunk_slice]
            buffers = self.data[chunk_slice]['data']

            chunk_processor = ChunkedBufferProcessor(buffer_ids, buffer_timestamps, buffers)
            chunk_processor.process_buffers()
            housekeeping = chunk_processor.merge_housekeeping()
            images_h, images_v = chunk_processor.merge_assembled_images()

            if self.name == '2DS':
                self.write_image_data(spiffile, group_h_channel, images_h)
                self.write_image_data(spiffile, group_v_channel, images_v)

                self.write_housekeeping_data(spiffile, group_h_channel, housekeeping)
                self.write_housekeeping_data(spiffile, group_v_channel, housekeeping)

            else:
                self.write_image_data(spiffile, group_no_channel, images_v)
                self.write_housekeeping_data(spiffile, group_no_channel, housekeeping)

            pbar1.update(chunk_intervals[i + 1] - chunk_intervals[i])

        tas_post_processor = SPECFileTASPostProcessor()

        if self.name == '2DS':
            tas_post_processor.post_process_tas(spiffile.rootgrp[group_h_channel])
            tas_post_processor.post_process_tas(spiffile.rootgrp[group_v_channel])
        else:
            tas_post_processor.post_process_tas(spiffile.rootgrp[group_no_channel])

    def get_chunk_intervals(self):
        data_len = len(self.data)

        chunk_intervals = np.arange(0, data_len, self.chunk_size)

        if chunk_intervals[-1] != data_len:
            chunk_intervals[-1] = data_len

        return chunk_intervals


    def write_image_data(self, spiffile, instrument, images):

        """ Called each time number of unsaved processed images exceeds
        10,000. Writes H and V images to SPIF NetCDF file.
        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        h_p : Images object
            Images object containing processed H array image info.
        v_p : Images object
            Images object containing processed V array image info.
        """

        instgrp = spiffile.rootgrp[instrument]
        coregrp = spiffile.rootgrp[instrument]['core']

        dim_size = len(instgrp.dimensions['Images'])
        px_size = len(instgrp.dimensions['Pixels'])

        # Primary, critical SPIF variables

        spiffile.write_variable(coregrp, 'image_sec', images.image_sec, dim_size)
        spiffile.write_variable(coregrp, 'image_ns', images.image_ns, dim_size)
        spiffile.write_variable(coregrp, 'image_len', images.image_len, dim_size)
        spiffile.write_variable(coregrp, 'buffer_index', images.buffer_id, dim_size)
        spiffile.write_variable(coregrp, 'image', images.image, px_size)

        # Auxiliary SPIF variables

        spiffile.write_variable(coregrp, 'num_words', images.num_words, dim_size)
        spiffile.write_variable(coregrp, 'timing_flag', images.timing_flag, dim_size)
        spiffile.write_variable(coregrp, 'mismatch_flag', images.mismatch_flag, dim_size)
        spiffile.write_variable(coregrp, 'fifo_flag', images.fifo_flag, dim_size)
        spiffile.write_variable(coregrp, 'overload_flag', images.overload_flag, dim_size)
        spiffile.write_variable(coregrp, 'particle_count', images.particle_count, dim_size)
        spiffile.write_variable(coregrp, 'num_slices', images.num_slices, dim_size)
        spiffile.write_variable(coregrp, 'timeword_upper', images.timeword_upper, dim_size)
        spiffile.write_variable(coregrp, 'timeword_lower', images.timeword_lower, dim_size)
        spiffile.write_variable(coregrp, 'tas', images.tas, dim_size)

        spiffile.rootgrp.sync()

    def write_housekeeping_data(self, spiffile, instrument, housekeeping):
        """
        Called each time number of unsaved processed images exceeds
        10,000. Writes H and V images to SPIF NetCDF file.
        Parameters
        ----------
        spiffile : SPIFFile object
            SPIFFile object of current SPIF NetCDF output file
        h_p : Images object
            Images object containing processed H array image info.
        v_p : Images object
            Images object containing processed V array image info.
        """

        instgrp = spiffile.rootgrp[instrument]
        hk_group = spiffile.rootgrp[instrument]['aux-housekeeping']

        dim_size = len(hk_group.dimensions['Packet'])

        # Primary, critical SPIF variables

        spiffile.write_variable( hk_group, "buffer_index",                  housekeeping["buffer_id"][:],                     dim_size )
        spiffile.write_variable( hk_group, "buffer_sec",                    housekeeping["buffer_sec"][:],                    dim_size )
        spiffile.write_variable( hk_group, "buffer_ns",                     housekeeping["buffer_ns"][:],                     dim_size )
        spiffile.write_variable( hk_group, "incomplete_packet",             housekeeping["incomplete_packet"][:],             dim_size )
        spiffile.write_variable( hk_group, "hz_elem_0_voltage",             housekeeping["hz_elem_0_voltage"][:],             dim_size )
        spiffile.write_variable( hk_group, "hz_elem_64_voltage",            housekeeping["hz_elem_64_voltage"][:],            dim_size )
        spiffile.write_variable( hk_group, "hz_elem_127_voltage",           housekeeping["hz_elem_127_voltage"][:],           dim_size )
        spiffile.write_variable( hk_group, "v_elem_0_voltage",              housekeeping["v_elem_0_voltage"][:],              dim_size )
        spiffile.write_variable( hk_group, "v_elem_64_voltage",             housekeeping["v_elem_64_voltage"][:],             dim_size )
        spiffile.write_variable( hk_group, "v_elem_127_voltage",            housekeeping["v_elem_127_voltage"][:],            dim_size )
        spiffile.write_variable( hk_group, "raw_pos_power_supply",          housekeeping["raw_pos_power_supply"][:],          dim_size )
        spiffile.write_variable( hk_group, "raw_neg_power_supply",          housekeeping["raw_neg_power_supply"][:],          dim_size )
        spiffile.write_variable( hk_group, "hz_arm_tx_temp",                housekeeping["hz_arm_tx_temp"][:],                dim_size )
        spiffile.write_variable( hk_group, "hz_arm_rx_temp",                housekeeping["hz_arm_rx_temp"][:],                dim_size )
        spiffile.write_variable( hk_group, "v_arm_tx_temp",                 housekeeping["v_arm_tx_temp"][:],                 dim_size )
        spiffile.write_variable( hk_group, "v_arm_rx_temp",                 housekeeping["v_arm_rx_temp"][:],                 dim_size )
        spiffile.write_variable( hk_group, "hz_tip_tx_temp",                housekeeping["hz_tip_tx_temp"][:],                dim_size )
        spiffile.write_variable( hk_group, "hz_tip_rx_temp",                housekeeping["hz_tip_rx_temp"][:],                dim_size )
        spiffile.write_variable( hk_group, "rear_opt_bridge_temp",          housekeeping["rear_opt_bridge_temp"][:],          dim_size )
        spiffile.write_variable( hk_group, "dsp_board_temp",                housekeeping["dsp_board_temp"][:],                dim_size )
        spiffile.write_variable( hk_group, "forward_vessel_temp",           housekeeping["forward_vessel_temp"][:],           dim_size )
        spiffile.write_variable( hk_group, "hz_laser_temp",                 housekeeping["hz_laser_temp"][:],                 dim_size )
        spiffile.write_variable( hk_group, "v_laser_temp",                  housekeeping["v_laser_temp"][:],                  dim_size )
        spiffile.write_variable( hk_group, "front_plate_temp",              housekeeping["front_plate_temp"][:],              dim_size )
        spiffile.write_variable( hk_group, "power_supply_temp",             housekeeping["power_supply_temp"][:],             dim_size )
        spiffile.write_variable( hk_group, "minus_5V_supply",               housekeeping["minus_5V_supply"][:],               dim_size )
        spiffile.write_variable( hk_group, "plus_5V_supply",                housekeeping["plus_5V_supply"][:],                dim_size )
        spiffile.write_variable( hk_group, "can_internal_pressure",         housekeeping["can_internal_pressure"][:],         dim_size )
        spiffile.write_variable( hk_group, "hz_elem_21_voltage",            housekeeping["hz_elem_21_voltage"][:],            dim_size )
        spiffile.write_variable( hk_group, "hz_elem_42_voltage",            housekeeping["hz_elem_42_voltage"][:],            dim_size )
        spiffile.write_variable( hk_group, "hz_elem_85_voltage",            housekeeping["hz_elem_85_voltage"][:],            dim_size )
        spiffile.write_variable( hk_group, "hz_elem_106_voltage",           housekeeping["hz_elem_106_voltage"][:],           dim_size )
        spiffile.write_variable( hk_group, "v_elem_21_voltage",             housekeeping["v_elem_21_voltage"][:],             dim_size )
        spiffile.write_variable( hk_group, "v_elem_42_voltage",             housekeeping["v_elem_42_voltage"][:],             dim_size )
        spiffile.write_variable( hk_group, "v_elem_85_voltage",             housekeeping["v_elem_85_voltage"][:],             dim_size )
        spiffile.write_variable( hk_group, "v_elem_106_voltage",            housekeeping["v_elem_106_voltage"][:],            dim_size )
        spiffile.write_variable( hk_group, "num_v_particles_detected",      housekeeping["num_v_particles_detected"][:],      dim_size )
        spiffile.write_variable( hk_group, "num_h_particles_detected",      housekeeping["num_h_particles_detected"][:],      dim_size )
        spiffile.write_variable( hk_group, "heater_outputs",                housekeeping["heater_outputs"][:],                dim_size )
        spiffile.write_variable( hk_group, "h_laser_drive",                 housekeeping["h_laser_drive"][:],                 dim_size )
        spiffile.write_variable( hk_group, "v_laser_drive",                 housekeeping["v_laser_drive"][:],                 dim_size )
        spiffile.write_variable( hk_group, "hz_masked_bits",                housekeeping["hz_masked_bits"][:],                dim_size )
        spiffile.write_variable( hk_group, "v_masked_bits",                 housekeeping["v_masked_bits"][:],                 dim_size )
        spiffile.write_variable( hk_group, "num_stereo_particles_detected", housekeeping["num_stereo_particles_detected"][:], dim_size )
        spiffile.write_variable( hk_group, "num_t_word_mismatch",           housekeeping["num_t_word_mismatch"][:],           dim_size )
        spiffile.write_variable( hk_group, "num_slice_count_mismatch",      housekeeping["num_slice_count_mismatch"][:],      dim_size )
        spiffile.write_variable( hk_group, "num_hz_ov_periods",             housekeeping["num_hz_ov_periods"][:],             dim_size )
        spiffile.write_variable( hk_group, "num_v_ov_periods",              housekeeping["num_v_ov_periods"][:],              dim_size )
        spiffile.write_variable( hk_group, "compression_conf",              housekeeping["compression_conf"][:],              dim_size )
        spiffile.write_variable( hk_group, "num_empty_fifo",                housekeeping["num_empty_fifo"][:],                dim_size )
        spiffile.write_variable( hk_group, "tas",                           housekeeping["tas"][:],                           dim_size )
        spiffile.write_variable( hk_group, "timing_word_1",                 housekeeping["timing_word_1"][:],                 dim_size )
        spiffile.write_variable( hk_group, "timing_word_2",                 housekeeping["timing_word_2"][:],                 dim_size )

        spiffile.rootgrp.sync()

class ChunkedBufferProcessor:

    def __init__(self, buffer_ids, buffer_timestamps, raw_buffers):
        self.buffer_ids = buffer_ids
        self.buffer_timestamps = buffer_timestamps
        self.raw_buffers = raw_buffers

        self.processed_buffers = [None]*len(raw_buffers)

        self.output = [None]*len(raw_buffers)


    def process_buffers(self):

        for i in range(len(self.output)):
            self.output[i] = self.process_buffer(
                self.buffer_ids[i],
                self.buffer_timestamps[i],
                self.raw_buffers[i],
                i
            )

    def process_buffer(self, buffer_id, buffer_timestamp, buffer, output_id):

        buffer = Buffer(buffer_id, buffer_timestamp, buffer)
        self.processed_buffers[output_id] = buffer

        return buffer.assembled_images

    def merge_assembled_images(self):

        image_container_h = AssembledImageRecordContainer()
        image_container_v = AssembledImageRecordContainer()

        for k in image_container_h.__dict__.keys():
            image_container_h.__dict__[k] = np.concatenate([x['h'].__dict__[k] for x in self.output])
            image_container_v.__dict__[k] = np.concatenate([x['v'].__dict__[k] for x in self.output])

        return image_container_h, image_container_v

    def merge_housekeeping(self):
        housekeeping_data = np.array([], dtype = housekeeping_template)

        for buffer in self.processed_buffers:
            housekeeping_data = np.append(housekeeping_data, buffer.housekeeping)

        housekeeping_data['housekeeping_packet_id'] = np.arange(0,len(housekeeping_data))

        return housekeeping_data

class SPECFileTASPostProcessor:

    def __init__(self) -> None:
        pass

    def post_process_tas(self, instfile):

        aux_hk_group = instfile["aux-housekeeping"]
        core_group = instfile["core"]

        self.clean_hk_tas(aux_hk_group)
        self.set_core_tas(aux_hk_group, core_group)

    def clean_hk_tas(self, aux_hk_group):
        tas_clean = aux_hk_group['tas'][:]

        first_clean_tas_index = np.argwhere( ~np.isnan(tas_clean) )[0]

        # do back filling first just in case

        if first_clean_tas_index != 0:
            tas_clean[0:first_clean_tas_index] = tas_clean[first_clean_tas_index]

        for i in range(1,len(tas_clean)):
            if np.isnan(tas_clean[i]):
                tas_clean[i] = tas_clean[i - 1]

        aux_hk_group['tas'][:] = tas_clean

    def set_core_tas(self, aux_hk_group, core_group):
        buffer_id_hk = aux_hk_group['buffer_index'][:]
        tas_hk = aux_hk_group['tas'][:]

        buffer_id_core = core_group['buffer_index'][:]
        tas_core = core_group['tas'][:]

        f = interp1d(buffer_id_hk, tas_hk, kind = 'previous', fill_value='extrapolate')

        interpolated_tas = f( np.arange(0,np.max(buffer_id_core) + 1) )

        updated_core_tas = interpolated_tas[buffer_id_core]

        core_group['tas'][:] = updated_core_tas
